import base64
from io import IOBase
from pathlib import Path
from typing import Any, List, Literal

import PIL.Image
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from pydantic import BaseModel, HttpUrl
from ..._resource import AsyncAPIResource, SyncAPIResource
from ...utils.ai_models import assert_valid_model_extraction
from ...utils.json_schema import load_json_schema
from ...utils.mime import MIMEData, prepare_mime_document
from ...types.browser_canvas import BrowserCanvas
from ...types.documents.extractions import RetabParsedChatCompletion
from ...types.logs import ProcessorConfig, UpdateProcessorRequest
from ...types.modalities import Modality
from ...types.pagination import ListMetadata
from ...types.standards import PreparedRequest, FieldUnset
from .automations.client import AsyncAutomations, Automations


class ListProcessors(BaseModel):
    """Response model for listing processor configurations."""

    data: list[ProcessorConfig]
    list_metadata: ListMetadata


class ProcessorsMixin:
    def prepare_create(
        self,
        name: str,
        json_schema: dict[str, Any] | Path | str,
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = FieldUnset,
        reasoning_effort: ChatCompletionReasoningEffort = FieldUnset,
        image_resolution_dpi: int = FieldUnset,
        browser_canvas: BrowserCanvas = FieldUnset,
        n_consensus: int = FieldUnset,
    ) -> PreparedRequest:
        assert_valid_model_extraction(model)

        # Load the JSON schema from file path, string, or dict
        loaded_schema = load_json_schema(json_schema)

        config_dict = {
            'name': name,
            'json_schema': loaded_schema,
            'modality': modality,
            'model': model,
        }
        if temperature is not FieldUnset:
            config_dict['temperature'] = temperature
        if reasoning_effort is not FieldUnset:
            config_dict['reasoning_effort'] = reasoning_effort
        if image_resolution_dpi is not FieldUnset:
            config_dict['image_resolution_dpi'] = image_resolution_dpi
        if browser_canvas is not FieldUnset:
            config_dict['browser_canvas'] = browser_canvas
        if n_consensus is not FieldUnset:
            config_dict['n_consensus'] = n_consensus

        processor_config = ProcessorConfig(**config_dict)

        return PreparedRequest(method="POST", url="/v1/processors", data=processor_config.model_dump(mode="json", exclude_unset=True))

    def prepare_list(
        self,
        before: str | None = None,
        after: str | None = None,
        limit: int | None = 10,
        order: Literal["asc", "desc"] | None = "desc",
        # Filtering parameters
        name: str | None = None,
        modality: str | None = None,
        model: str | None = None,
        schema_id: str | None = None,
        schema_data_id: str | None = None,
    ) -> PreparedRequest:
        params = {
            "before": before,
            "after": after,
            "limit": limit,
            "order": order,
            "name": name,
            "modality": modality,
            "model": model,
            "schema_id": schema_id,
            "schema_data_id": schema_data_id,
        }
        # Remove None values
        params = {k: v for k, v in params.items() if v is not None}

        return PreparedRequest(method="GET", url="/v1/processors", params=params)

    def prepare_get(self, processor_id: str) -> PreparedRequest:
        """Get a specific processor configuration.

        Args:
            processor_id: ID of the processor

        Returns:
            ProcessorConfig: The processor configuration
        """
        return PreparedRequest(method="GET", url=f"/v1/processors/{processor_id}")

    def prepare_update(
        self,
        processor_id: str,
        name: str | None = None,
        modality: Modality | None = None,
        image_resolution_dpi: int | None = None,
        browser_canvas: BrowserCanvas | None = None,
        model: str | None = None,
        json_schema: dict[str, Any] | Path | str | None = None,
        temperature: float | None = None,
        reasoning_effort: ChatCompletionReasoningEffort | None = None,
        n_consensus: int | None = None,
    ) -> PreparedRequest:
        if model is not None:
            assert_valid_model_extraction(model)

        # Load the JSON schema from file path, string, or dict if provided
        loaded_schema = None
        if json_schema is not None:
            loaded_schema = load_json_schema(json_schema)

        update_request = UpdateProcessorRequest(
            name=name,
            modality=modality,
            image_resolution_dpi=image_resolution_dpi,
            browser_canvas=browser_canvas,
            model=model,
            json_schema=loaded_schema,
            temperature=temperature,
            reasoning_effort=reasoning_effort,
            n_consensus=n_consensus,
        )

        return PreparedRequest(method="PUT", url=f"/v1/processors/{processor_id}", data=update_request.model_dump(mode="json", exclude_none=True))

    def prepare_delete(self, processor_id: str) -> PreparedRequest:
        return PreparedRequest(method="DELETE", url=f"/v1/processors/{processor_id}")

    def prepare_submit(
        self,
        processor_id: str,
        document: Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl | None = None,
        documents: list[Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl] | None = None,
        temperature: float | None = None,
        seed: int | None = None,
        store: bool = True,
    ) -> PreparedRequest:
        """Prepare a request to submit documents to a processor.

        Args:
            processor_id: ID of the processor
            document: Single document to process (mutually exclusive with documents)
            documents: List of documents to process (mutually exclusive with document)
            temperature: Optional temperature override
            seed: Optional seed for reproducibility
            store: Whether to store the results

        Returns:
            PreparedRequest: The prepared request
        """
        # Validate that either document or documents is provided, but not both
        if not document and not documents:
            raise ValueError("Either 'document' or 'documents' must be provided")

        if document and documents:
            raise ValueError("Provide either 'document' (single) or 'documents' (multiple), not both")

        # Prepare form data parameters
        form_data = {
            "temperature": temperature,
            "seed": seed,
            "store": store,
        }
        # Remove None values
        form_data = {k: v for k, v in form_data.items() if v is not None}

        # Prepare files for upload
        files = {}
        if document:
            # Convert document to MIMEData if needed
            mime_document = prepare_mime_document(document)
            # Single document upload
            files["document"] = (mime_document.filename, base64.b64decode(mime_document.content), mime_document.mime_type)
        elif documents:
            # Multiple documents upload - httpx supports multiple files with same field name using a list
            files_list = []
            for doc in documents:
                # Convert each document to MIMEData if needed
                mime_doc = prepare_mime_document(doc)
                files_list.append(
                    (
                        "documents",  # field name
                        (mime_doc.filename, base64.b64decode(mime_doc.content), mime_doc.mime_type),
                    )
                )
            files = files_list

        url = f"/v1/processors/{processor_id}/submit"

        return PreparedRequest(method="POST", url=url, form_data=form_data, files=files)


class Processors(SyncAPIResource, ProcessorsMixin):
    """Processors API wrapper for managing processor configurations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.automations = Automations(client=client)

    def create(
        self,
        name: str,
        json_schema: dict[str, Any] | Path | str,
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = FieldUnset,
        reasoning_effort: ChatCompletionReasoningEffort = FieldUnset,
        image_resolution_dpi: int = FieldUnset,
        browser_canvas: BrowserCanvas = FieldUnset,
        n_consensus: int = FieldUnset,
    ) -> ProcessorConfig:
        """Create a new processor configuration.

        Args:
            name: Name of the processor
            json_schema: JSON schema for the processor. Can be a dictionary, file path (Path or str), or JSON string.
            image_resolution_dpi: Optional image resolution DPI
            browser_canvas: Optional browser canvas size
            modality: Processing modality (currently only "native" supported)
            model: AI model to use for processing
            temperature: Model temperature setting
            reasoning_effort: The effort level for the model to reason about the input data.
            n_consensus: Number of consensus required to validate the data
        Returns:
            ProcessorConfig: The created processor configuration
        """
        request = self.prepare_create(
            name=name,
            json_schema=json_schema,
            modality=modality,
            model=model,
            temperature=temperature,
            reasoning_effort=reasoning_effort,
            image_resolution_dpi=image_resolution_dpi,
            browser_canvas=browser_canvas,
            n_consensus=n_consensus,
        )
        response = self._client._prepared_request(request)
        print(f"Processor ID: {response['id']}. Processor available at https://www.retab.com/dashboard/processors/{response['id']}")
        return ProcessorConfig.model_validate(response)

    def list(
        self,
        before: str | None = None,
        after: str | None = None,
        limit: int | None = 10,
        order: Literal["asc", "desc"] | None = "desc",
        name: str | None = None,
        modality: str | None = None,
        model: str | None = None,
        schema_id: str | None = None,
        schema_data_id: str | None = None,
    ) -> ListProcessors:
        """List processor configurations with pagination support.

        Args:
            before: Optional cursor for pagination before a specific processor ID
            after: Optional cursor for pagination after a specific processor ID
            limit: Optional limit on number of results (max 100)
            order: Optional sort order ("asc" or "desc")
            name: Optional filter by processor name
            modality: Optional filter by modality
            model: Optional filter by model
            schema_id: Optional filter by schema ID
            schema_data_id: Optional filter by schema data ID

        Returns:
            ListProcessors: Paginated list of processor configurations with metadata
        """
        request = self.prepare_list(before, after, limit, order, name, modality, model, schema_id, schema_data_id)
        response = self._client._prepared_request(request)
        return ListProcessors.model_validate(response)

    def get(self, processor_id: str) -> ProcessorConfig:
        """Get a specific processor configuration.

        Args:
            processor_id: ID of the processor

        Returns:
            ProcessorConfig: The processor configuration
        """
        request = self.prepare_get(processor_id)
        response = self._client._prepared_request(request)
        return ProcessorConfig.model_validate(response)

    def update(
        self,
        processor_id: str,
        name: str | None = None,
        modality: Modality | None = None,
        image_resolution_dpi: int | None = None,
        browser_canvas: BrowserCanvas | None = None,
        model: str | None = None,
        json_schema: dict[str, Any] | Path | str | None = None,
        temperature: float | None = None,
        reasoning_effort: ChatCompletionReasoningEffort | None = None,
        n_consensus: int | None = None,
    ) -> ProcessorConfig:
        """Update a processor configuration.

        Args:
            processor_id: ID of the processor to update
            name: New name for the processor
            modality: New processing modality
            image_resolution_dpi: New image resolution DPI
            browser_canvas: New browser canvas size
            model: New AI model
            json_schema: New JSON schema for the processor. Can be a dictionary, file path (Path or str), or JSON string.
            temperature: New temperature setting
            reasoning_effort: The effort level for the model to reason about the input data.
            n_consensus: New number of consensus required
        Returns:
            ProcessorConfig: The updated processor configuration
        """
        request = self.prepare_update(
            processor_id=processor_id,
            name=name,
            modality=modality,
            image_resolution_dpi=image_resolution_dpi,
            browser_canvas=browser_canvas,
            model=model,
            json_schema=json_schema,
            temperature=temperature,
            reasoning_effort=reasoning_effort,
            n_consensus=n_consensus,
        )
        response = self._client._prepared_request(request)
        return ProcessorConfig.model_validate(response)

    def delete(self, processor_id: str) -> None:
        """Delete a processor configuration.

        Args:
            processor_id: ID of the processor to delete
        """
        request = self.prepare_delete(processor_id)
        self._client._prepared_request(request)
        print(f"Processor Deleted. ID: {processor_id}")

    def submit(
        self,
        processor_id: str,
        document: Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl | None = None,
        documents: List[Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl] | None = None,
        temperature: float | None = None,
        seed: int | None = None,
        store: bool = True,
    ) -> RetabParsedChatCompletion:
        """Submit documents to a processor for processing.

        Args:
            processor_id: ID of the processor
            document: Single document to process (mutually exclusive with documents)
            documents: List of documents to process (mutually exclusive with document)
            temperature: Optional temperature override
            seed: Optional seed for reproducibility
            store: Whether to store the results

        Returns:
            RetabParsedChatCompletion: The processing result
        """
        request = self.prepare_submit(processor_id=processor_id, document=document, documents=documents, temperature=temperature, seed=seed, store=store)
        response = self._client._prepared_request(request)
        return RetabParsedChatCompletion.model_validate(response)


class AsyncProcessors(AsyncAPIResource, ProcessorsMixin):
    """Async Processors API wrapper for managing processor configurations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.automations = AsyncAutomations(client=client)

    async def create(
        self,
        name: str,
        json_schema: dict[str, Any] | Path | str,
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = FieldUnset,
        reasoning_effort: ChatCompletionReasoningEffort = FieldUnset,
        image_resolution_dpi: int = FieldUnset,
        browser_canvas: BrowserCanvas = FieldUnset,
        n_consensus: int = FieldUnset,
    ) -> ProcessorConfig:
        request = self.prepare_create(
            name=name,
            json_schema=json_schema,
            modality=modality,
            model=model,
            temperature=temperature,
            reasoning_effort=reasoning_effort,
            image_resolution_dpi=image_resolution_dpi,
            browser_canvas=browser_canvas,
            n_consensus=n_consensus,
        )
        response = await self._client._prepared_request(request)
        print(f"Processor ID: {response['id']}. Processor available at https://www.retab.com/dashboard/processors/{response['id']}")

        return ProcessorConfig.model_validate(response)

    async def list(
        self,
        before: str | None = None,
        after: str | None = None,
        limit: int | None = 10,
        order: Literal["asc", "desc"] | None = "desc",
        name: str | None = None,
        modality: str | None = None,
        model: str | None = None,
        schema_id: str | None = None,
        schema_data_id: str | None = None,
    ) -> ListProcessors:
        request = self.prepare_list(before, after, limit, order, name, modality, model, schema_id, schema_data_id)
        response = await self._client._prepared_request(request)
        return ListProcessors.model_validate(response)

    async def get(self, processor_id: str) -> ProcessorConfig:
        request = self.prepare_get(processor_id)
        response = await self._client._prepared_request(request)
        return ProcessorConfig.model_validate(response)

    async def update(
        self,
        processor_id: str,
        name: str | None = None,
        modality: Modality | None = None,
        image_resolution_dpi: int | None = None,
        browser_canvas: BrowserCanvas | None = None,
        model: str | None = None,
        json_schema: dict[str, Any] | Path | str | None = None,
        temperature: float | None = None,
        reasoning_effort: ChatCompletionReasoningEffort | None = None,
        n_consensus: int | None = None,
    ) -> ProcessorConfig:
        """Update a processor configuration.

        Args:
            processor_id: ID of the processor to update
            name: New name for the processor
            modality: New processing modality
            image_resolution_dpi: New image resolution DPI
            browser_canvas: New browser canvas size
            model: New AI model
            json_schema: New JSON schema for the processor. Can be a dictionary, file path (Path or str), or JSON string.
            temperature: New temperature setting
            reasoning_effort: The effort level for the model to reason about the input data.
            n_consensus: New number of consensus required
        Returns:
            ProcessorConfig: The updated processor configuration
        """
        request = self.prepare_update(processor_id, name, modality, image_resolution_dpi, browser_canvas, model, json_schema, temperature, reasoning_effort, n_consensus)
        response = await self._client._prepared_request(request)
        return ProcessorConfig.model_validate(response)

    async def delete(self, processor_id: str) -> None:
        request = self.prepare_delete(processor_id)
        await self._client._prepared_request(request)
        print(f"Processor Deleted. ID: {processor_id}")

    async def submit(
        self,
        processor_id: str,
        document: Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl | None = None,
        documents: List[Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl] | None = None,
        temperature: float | None = None,
        seed: int | None = None,
        store: bool = True,
    ) -> RetabParsedChatCompletion:
        """Submit documents to a processor for processing.

        Args:
            processor_id: ID of the processor
            document: Single document to process (mutually exclusive with documents)
            documents: List of documents to process (mutually exclusive with document)
            temperature: Optional temperature override
            seed: Optional seed for reproducibility
            store: Whether to store the results

        Returns:
            RetabParsedChatCompletion: The processing result
        """
        request = self.prepare_submit(processor_id=processor_id, document=document, documents=documents, temperature=temperature, seed=seed, store=store)
        response = await self._client._prepared_request(request)
        return RetabParsedChatCompletion.model_validate(response)



import hashlib
import hmac
import json
from typing import Any, Literal, Optional, Union

from ...._resource import AsyncAPIResource, SyncAPIResource
from ....types.automations.endpoints import Endpoint, UpdateEndpointRequest
from ....types.automations.links import Link, UpdateLinkRequest
from ....types.automations.mailboxes import Mailbox, UpdateMailboxRequest
from ....types.automations.outlook import Outlook, UpdateOutlookRequest
from ....types.standards import PreparedRequest
from .endpoints import AsyncEndpoints, Endpoints
from .links import AsyncLinks, Links
from .logs import AsyncLogs, Logs
from .mailboxes import AsyncMailboxes, Mailboxes
from .outlook import AsyncOutlooks, Outlooks
from .tests import AsyncTests, Tests


class SignatureVerificationError(Exception):
    """Raised when webhook signature verification fails."""

    pass


class AutomationsMixin:
    def _verify_event(self, event_body: bytes, event_signature: str, secret: str) -> Any:
        """
        Verify the signature of a webhook event.

        Args:
            body: The raw request body
            signature: The signature header
            secret: The secret key used for signing

        Returns:
            The parsed event payload

        Raises:
            SignatureVerificationError: If the signature verification fails
        """
        expected_signature = hmac.new(secret.encode(), event_body, hashlib.sha256).hexdigest()

        if not hmac.compare_digest(event_signature, expected_signature):
            raise SignatureVerificationError("Invalid signature")

        return json.loads(event_body.decode("utf-8"))

    def prepare_list(
        self,
        processor_id: str,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        automation_id: Optional[str] = None,
        webhook_url: Optional[str] = None,
        name: Optional[str] = None,
    ) -> PreparedRequest:
        params = {
            "before": before,
            "after": after,
            "limit": limit,
            "order": order,
            "id": automation_id,
            "webhook_url": webhook_url,
            "name": name,
        }
        params = {k: v for k, v in params.items() if v is not None}
        return PreparedRequest(method="GET", url=f"/v1/processors/{processor_id}/automations", params=params)

    def prepare_get(self, processor_id: str, automation_id: str) -> PreparedRequest:
        return PreparedRequest(method="GET", url=f"/v1/processors/{processor_id}/automations/{automation_id}")

    def prepare_update(
        self,
        processor_id: str,
        automation_id: str,
        automation_data: Union[UpdateLinkRequest, UpdateMailboxRequest, UpdateEndpointRequest, UpdateOutlookRequest],
    ) -> PreparedRequest:
        return PreparedRequest(method="PUT", url=f"/v1/processors/{processor_id}/automations/{automation_id}", data=automation_data.model_dump(mode="json"))

    def prepare_delete(self, processor_id: str, automation_id: str) -> PreparedRequest:
        return PreparedRequest(method="DELETE", url=f"/v1/processors/{processor_id}/automations/{automation_id}")


class Automations(SyncAPIResource, AutomationsMixin):
    """Automations API wrapper"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.mailboxes = Mailboxes(client=client)
        self.links = Links(client=client)
        self.outlook = Outlooks(client=client)
        self.endpoints = Endpoints(client=client)
        self.tests = Tests(client=client)
        self.logs = Logs(client=client)

    def verify_event(self, event_body: bytes, event_signature: str, secret: str) -> Any:
        """
        Verify the signature of a webhook event.
        """
        return self._verify_event(event_body, event_signature, secret)

    def list_automations(
        self,
        processor_id: str,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        automation_id: Optional[str] = None,
        webhook_url: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """List automations attached to this processor."""
        request = self.prepare_list(processor_id, before, after, limit, order, automation_id, webhook_url, name)
        response = self._client._prepared_request(request)
        return response

    def get_automation(self, processor_id: str, automation_id: str) -> Union[Link, Mailbox, Endpoint, Outlook]:
        """Get a specific automation attached to this processor."""
        request = self.prepare_get(processor_id, automation_id)
        response = self._client._prepared_request(request)

        # Return the appropriate model based on the automation type
        if response["object"] == "automation.link":
            return Link.model_validate(response)
        elif response["object"] == "automation.mailbox":
            return Mailbox.model_validate(response)
        elif response["object"] == "automation.endpoint":
            return Endpoint.model_validate(response)
        elif response["object"] == "automation.outlook":
            return Outlook.model_validate(response)
        else:
            raise ValueError(f"Unknown automation type: {response.get('object')}")

    def update_automation(
        self,
        processor_id: str,
        automation_id: str,
        automation_data: Union[UpdateLinkRequest, UpdateMailboxRequest, UpdateEndpointRequest, UpdateOutlookRequest],
    ) -> Union[Link, Mailbox, Endpoint, Outlook]:
        """Update an automation attached to this processor."""
        request = self.prepare_update(processor_id, automation_id, automation_data)
        response = self._client._prepared_request(request)

        # Return the appropriate model based on the automation type
        if response["object"] == "automation.link":
            return Link.model_validate(response)
        elif response["object"] == "automation.mailbox":
            return Mailbox.model_validate(response)
        elif response["object"] == "automation.endpoint":
            return Endpoint.model_validate(response)
        elif response["object"] == "automation.outlook":
            return Outlook.model_validate(response)
        else:
            raise ValueError(f"Unknown automation type: {response.get('object')}")

    def delete_automation(self, processor_id: str, automation_id: str) -> None:
        """Delete an automation attached to this processor."""
        request = self.prepare_delete(processor_id, automation_id)
        self._client._prepared_request(request)
        print(f"Automation {automation_id} deleted from processor {processor_id}")


class AsyncAutomations(AsyncAPIResource, AutomationsMixin):
    """Async Automations API wrapper"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.mailboxes = AsyncMailboxes(client=client)
        self.links = AsyncLinks(client=client)
        self.outlook = AsyncOutlooks(client=client)
        self.endpoints = AsyncEndpoints(client=client)
        self.tests = AsyncTests(client=client)
        self.logs = AsyncLogs(client=client)

    async def verify_event(self, event_body: bytes, event_signature: str, secret: str) -> Any:
        """
        Verify the signature of a webhook event.
        """
        return self._verify_event(event_body, event_signature, secret)

    async def list(
        self,
        processor_id: str,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        automation_id: Optional[str] = None,
        webhook_url: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """List automations attached to this processor."""
        request = self.prepare_list(processor_id, before, after, limit, order, automation_id, webhook_url, name)
        response = await self._client._prepared_request(request)
        return response

    async def get(self, processor_id: str, automation_id: str) -> Union[Link, Mailbox, Endpoint, Outlook]:
        """Get a specific automation attached to this processor."""
        request = self.prepare_get(processor_id, automation_id)
        response = await self._client._prepared_request(request)

        # Return the appropriate model based on the automation type
        if response["object"] == "automation.link":
            return Link.model_validate(response)
        elif response["object"] == "automation.mailbox":
            return Mailbox.model_validate(response)
        elif response["object"] == "automation.endpoint":
            return Endpoint.model_validate(response)
        elif response["object"] == "automation.outlook":
            return Outlook.model_validate(response)
        else:
            raise ValueError(f"Unknown automation type: {response.get('object')}")

    async def update(
        self,
        processor_id: str,
        automation_id: str,
        automation_data: Union[UpdateLinkRequest, UpdateMailboxRequest, UpdateEndpointRequest, UpdateOutlookRequest],
    ) -> Union[Link, Mailbox, Endpoint, Outlook]:
        """Update an automation attached to this processor."""
        request = self.prepare_update(processor_id, automation_id, automation_data)
        response = await self._client._prepared_request(request)

        # Return the appropriate model based on the automation type
        if response["object"] == "automation.link":
            return Link.model_validate(response)
        elif response["object"] == "automation.mailbox":
            return Mailbox.model_validate(response)
        elif response["object"] == "automation.endpoint":
            return Endpoint.model_validate(response)
        elif response["object"] == "automation.outlook":
            return Outlook.model_validate(response)
        else:
            raise ValueError(f"Unknown automation type: {response.get('object')}")

    async def delete(self, processor_id: str, automation_id: str) -> None:
        """Delete an automation attached to this processor."""
        request = self.prepare_delete(processor_id, automation_id)
        await self._client._prepared_request(request)
        print(f"Automation {automation_id} deleted from processor {processor_id}")



from typing import Any, Dict, List, Literal, Optional

from ..._resource import AsyncAPIResource, SyncAPIResource
from ...types.consensus import ReconciliationResponse, ReconciliationRequest
from ...types.standards import PreparedRequest
from .completions import AsyncCompletions, Completions
from .responses import AsyncResponses, Responses


class BaseConsensusMixin:
    def _prepare_reconcile(
        self,
        list_dicts: List[Dict[str, Any]],
        reference_schema: Optional[Dict[str, Any]] = None,
        mode: Literal["direct", "aligned"] = "direct",
        idempotency_key: str | None = None,
    ) -> PreparedRequest:
        request = ReconciliationRequest(
            list_dicts=list_dicts,
            reference_schema=reference_schema,
            mode=mode,
        )

        return PreparedRequest(method="POST", url="/v1/consensus/reconcile", data=request.model_dump(mode="json", exclude_unset=True), idempotency_key=idempotency_key)


class Consensus(SyncAPIResource, BaseConsensusMixin):
    """Consensus API wrapper for synchronous operations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.completions = Completions(client=client)
        self.responses = Responses(client=client)

    def reconcile(
        self,
        list_dicts: List[Dict[str, Any]],
        reference_schema: Optional[Dict[str, Any]] = None,
        mode: Literal["direct", "aligned"] = "direct",
        idempotency_key: str | None = None,
    ) -> ReconciliationResponse:
        """
        Reconcile multiple dictionaries to produce a single unified consensus dictionary.

        Args:
            list_dicts: List of dictionaries to reconcile
            reference_schema: Optional schema to validate dictionaries against
            mode: Mode for consensus computation ("direct" or "aligned")
            idempotency_key: Optional idempotency key for the request

        Returns:
            Dict containing the consensus dictionary and consensus likelihoods

        Raises:
            RetabAPIError: If the API request fails
        """
        request = self._prepare_reconcile(list_dicts, reference_schema, mode, idempotency_key)
        response = self._client._prepared_request(request)
        return ReconciliationResponse.model_validate(response)


class AsyncConsensus(AsyncAPIResource, BaseConsensusMixin):
    """Consensus API wrapper for asynchronous operations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.completions = AsyncCompletions(client=client)
        self.responses = AsyncResponses(client=client)

    async def reconcile(
        self,
        list_dicts: List[Dict[str, Any]],
        reference_schema: Optional[Dict[str, Any]] = None,
        mode: Literal["direct", "aligned"] = "direct",
        idempotency_key: str | None = None,
    ) -> ReconciliationResponse:
        """
        Reconcile multiple dictionaries to produce a single unified consensus dictionary asynchronously.

        Args:
            list_dicts: List of dictionaries to reconcile
            reference_schema: Optional schema to validate dictionaries against
            mode: Mode for consensus computation ("direct" or "aligned")
            idempotency_key: Optional idempotency key for the request

        Returns:
            Dict containing the consensus dictionary and consensus likelihoods

        Raises:
            RetabAPIError: If the API request fails
        """
        request = self._prepare_reconcile(list_dicts, reference_schema, mode, idempotency_key)
        response = await self._client._prepared_request(request)

        return ReconciliationResponse.model_validate(response)



import json
from typing import AsyncGenerator, Generator

from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from openai.types.chat.parsed_chat_completion import ParsedChatCompletionMessage
from openai.types.shared_params.response_format_json_schema import ResponseFormatJSONSchema

# from openai.lib._parsing import ResponseFormatT
from pydantic import BaseModel as ResponseFormatT

from ..._resource import AsyncAPIResource, SyncAPIResource
from ...utils.ai_models import assert_valid_model_extraction
from ...utils.json_schema import unflatten_dict
from ...utils.stream_context_managers import as_async_context_manager, as_context_manager
from ...types.chat import ChatCompletionRetabMessage
from ...types.completions import RetabChatCompletionsRequest
from ...types.documents.extractions import RetabParsedChatCompletion, RetabParsedChatCompletionChunk, RetabParsedChoice
from ...types.schemas.object import Schema
from ...types.standards import PreparedRequest


class BaseCompletionsMixin:
    def prepare_parse(
        self,
        response_format: type[ResponseFormatT],
        messages: list[ChatCompletionRetabMessage],
        model: str,
        temperature: float,
        reasoning_effort: ChatCompletionReasoningEffort,
        stream: bool,
        n_consensus: int,
        idempotency_key: str | None = None,
    ) -> PreparedRequest:
        assert_valid_model_extraction(model)

        json_schema = response_format.model_json_schema()
        schema_obj = Schema(json_schema=json_schema)

        request = RetabChatCompletionsRequest(
            messages=messages,
            response_format={
                "type": "json_schema",
                "json_schema": {
                    "name": schema_obj.id,
                    "schema": schema_obj.inference_json_schema,
                    "strict": True,
                },
            },
            model=model,
            temperature=temperature,
            stream=stream,
            reasoning_effort=reasoning_effort,
            n_consensus=n_consensus,
        )

        return PreparedRequest(method="POST", url="/v1/completions", data=request.model_dump(), idempotency_key=idempotency_key)

    def prepare_create(
        self,
        response_format: ResponseFormatJSONSchema,
        messages: list[ChatCompletionRetabMessage],
        model: str,
        temperature: float,
        reasoning_effort: ChatCompletionReasoningEffort,
        stream: bool,
        n_consensus: int,
        idempotency_key: str | None = None,
    ) -> PreparedRequest:
        json_schema = response_format["json_schema"].get("schema")

        assert isinstance(json_schema, dict), f"json_schema must be a dictionary, got {type(json_schema)}"

        schema_obj = Schema(json_schema=json_schema)

        # Validate DocumentAPIRequest data (raises exception if invalid)
        request = RetabChatCompletionsRequest(
            model=model,
            messages=messages,
            response_format={
                "type": "json_schema",
                "json_schema": {
                    "name": schema_obj.id,
                    "schema": schema_obj.inference_json_schema,
                    "strict": True,
                },
            },
            temperature=temperature,
            stream=stream,
            reasoning_effort=reasoning_effort,
            n_consensus=n_consensus,
        )

        return PreparedRequest(method="POST", url="/v1/completions", data=request.model_dump(), idempotency_key=idempotency_key)


class Completions(SyncAPIResource, BaseCompletionsMixin):
    """Multi-provider Completions API wrapper"""

    @as_context_manager
    def stream(
        self,
        response_format: type[ResponseFormatT],
        messages: list[ChatCompletionRetabMessage],
        model: str = "gpt-4o-2024-08-06",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        n_consensus: int = 1,
        idempotency_key: str | None = None,
    ) -> Generator[RetabParsedChatCompletion, None, None]:
        """
        Process messages using the Retab API with streaming enabled.

        Args:
            response_format: JSON schema defining the expected data structure
            messages: List of chat messages to parse
            model: The AI model to use for processing
            temperature: Model temperature setting (0-1)
            reasoning_effort: The effort level for the model to reason about the input data
            idempotency_key: Idempotency key for request

        Returns:
            Generator[RetabParsedChatCompletion]: Stream of parsed responses

        Usage:
        ```python
        with retab.completions.stream(json_schema, messages, model, temperature, reasoning_effort) as stream:
            for response in stream:
                print(response)
        ```
        """
        request = self.prepare_parse(
            response_format=response_format,
            messages=messages,
            model=model,
            temperature=temperature,
            reasoning_effort=reasoning_effort,
            stream=True,
            n_consensus=n_consensus,
            idempotency_key=idempotency_key,
        )

        # Request the stream and return a context manager
        ui_parsed_chat_completion_cum_chunk: RetabParsedChatCompletionChunk | None = None
        # Initialize the RetabParsedChatCompletion object
        ui_parsed_completion: RetabParsedChatCompletion = RetabParsedChatCompletion(
            id="",
            created=0,
            model="",
            object="chat.completion",
            likelihoods={},
            choices=[
                RetabParsedChoice(
                    index=0,
                    message=ParsedChatCompletionMessage(content="", role="assistant"),
                    finish_reason=None,
                    logprobs=None,
                )
            ],
        )
        for chunk_json in self._client._prepared_request_stream(request):
            if not chunk_json:
                continue
            ui_parsed_chat_completion_cum_chunk = RetabParsedChatCompletionChunk.model_validate(chunk_json).chunk_accumulator(ui_parsed_chat_completion_cum_chunk)
            # Basic stuff
            ui_parsed_completion.id = ui_parsed_chat_completion_cum_chunk.id
            ui_parsed_completion.created = ui_parsed_chat_completion_cum_chunk.created
            ui_parsed_completion.model = ui_parsed_chat_completion_cum_chunk.model

            # Update the ui_parsed_completion object
            ui_parsed_completion.likelihoods = unflatten_dict(ui_parsed_chat_completion_cum_chunk.choices[0].delta.flat_likelihoods)
            parsed = unflatten_dict(ui_parsed_chat_completion_cum_chunk.choices[0].delta.flat_parsed)
            ui_parsed_completion.choices[0].message.content = json.dumps(parsed)
            ui_parsed_completion.choices[0].message.parsed = parsed

            yield ui_parsed_completion

        # change the finish_reason to stop
        ui_parsed_completion.choices[0].finish_reason = "stop"
        yield ui_parsed_completion


class AsyncCompletions(AsyncAPIResource, BaseCompletionsMixin):
    """Multi-provider Completions API wrapper for asynchronous usage."""

    @as_async_context_manager
    async def stream(
        self,
        response_format: type[ResponseFormatT],
        messages: list[ChatCompletionRetabMessage],
        model: str = "gpt-4o-2024-08-06",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        n_consensus: int = 1,
        idempotency_key: str | None = None,
    ) -> AsyncGenerator[RetabParsedChatCompletion, None]:
        """
        Parse messages using the Retab API asynchronously with streaming.

        Args:
            json_schema: JSON schema defining the expected data structure
            messages: List of chat messages to parse
            model: The AI model to use
            temperature: Model temperature setting (0-1)
            reasoning_effort: The effort level for the model to reason about the input data
            n_consensus: Number of consensus models to use for extraction
            idempotency_key: Idempotency key for request

        Returns:
            AsyncGenerator[RetabParsedChatCompletion]: Stream of parsed responses

        Usage:
        ```python
        async with retab.completions.stream(json_schema, messages, model, temperature, reasoning_effort, n_consensus) as stream:
            async for response in stream:
                print(response)
        ```
        """
        request = self.prepare_parse(
            response_format=response_format,
            messages=messages,
            model=model,
            temperature=temperature,
            reasoning_effort=reasoning_effort,
            stream=True,
            n_consensus=n_consensus,
            idempotency_key=idempotency_key,
        )

        # Request the stream and return a context manager
        ui_parsed_chat_completion_cum_chunk: RetabParsedChatCompletionChunk | None = None
        # Initialize the RetabParsedChatCompletion object
        ui_parsed_completion: RetabParsedChatCompletion = RetabParsedChatCompletion(
            id="",
            created=0,
            model="",
            object="chat.completion",
            likelihoods={},
            choices=[
                RetabParsedChoice(
                    index=0,
                    message=ParsedChatCompletionMessage(content="", role="assistant"),
                    finish_reason=None,
                    logprobs=None,
                )
            ],
        )
        async for chunk_json in self._client._prepared_request_stream(request):
            if not chunk_json:
                continue
            ui_parsed_chat_completion_cum_chunk = RetabParsedChatCompletionChunk.model_validate(chunk_json).chunk_accumulator(ui_parsed_chat_completion_cum_chunk)
            # Basic stuff
            ui_parsed_completion.id = ui_parsed_chat_completion_cum_chunk.id
            ui_parsed_completion.created = ui_parsed_chat_completion_cum_chunk.created
            ui_parsed_completion.model = ui_parsed_chat_completion_cum_chunk.model

            # Update the ui_parsed_completion object
            ui_parsed_completion.likelihoods = unflatten_dict(ui_parsed_chat_completion_cum_chunk.choices[0].delta.flat_likelihoods)
            parsed = unflatten_dict(ui_parsed_chat_completion_cum_chunk.choices[0].delta.flat_parsed)
            ui_parsed_completion.choices[0].message.content = json.dumps(parsed)
            ui_parsed_completion.choices[0].message.parsed = parsed

            yield ui_parsed_completion

        # change the finish_reason to stop
        ui_parsed_completion.choices[0].finish_reason = "stop"
        yield ui_parsed_completion



import base64
from io import IOBase
from pathlib import Path
from typing import Any, List

import PIL.Image
from pydantic import HttpUrl
from ..._resource import AsyncAPIResource, SyncAPIResource
from ...utils.mime import MIMEData, prepare_mime_document
from ...types.documents.extractions import RetabParsedChatCompletion
from ...types.standards import PreparedRequest


class DeploymentsMixin:
    def prepare_extract(
        self,
        project_id: str,
        iteration_id: str,
        document: Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl | None = None,
        documents: list[Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl] | None = None,
        temperature: float | None = None,
        seed: int | None = None,
        store: bool = True,
    ) -> PreparedRequest:
        """Prepare a request to extract documents from a deployment.

        Args:
            project_id: ID of the project
            iteration_id: ID of the iteration
            document: Single document to process (mutually exclusive with documents)
            documents: List of documents to process (mutually exclusive with document)
            temperature: Optional temperature override
            seed: Optional seed for reproducibility
            store: Whether to store the results

        Returns:
            PreparedRequest: The prepared request
        """
        # Validate that either document or documents is provided, but not both
        if not document and not documents:
            raise ValueError("Either 'document' or 'documents' must be provided")

        if document and documents:
            raise ValueError("Provide either 'document' (single) or 'documents' (multiple), not both")

        # Prepare form data parameters
        form_data = {
            "temperature": temperature,
            "seed": seed,
            "store": store,
        }
        # Remove None values
        form_data = {k: v for k, v in form_data.items() if v is not None}

        # Prepare files for upload
        files = {}
        if document:
            # Convert document to MIMEData if needed
            mime_document = prepare_mime_document(document)
            # Single document upload
            files["document"] = (mime_document.filename, base64.b64decode(mime_document.content), mime_document.mime_type)
        elif documents:
            # Multiple documents upload - httpx supports multiple files with same field name using a list
            files_list = []
            for doc in documents:
                # Convert each document to MIMEData if needed
                mime_doc = prepare_mime_document(doc)
                files_list.append(
                    (
                        "documents",  # field name
                        (mime_doc.filename, base64.b64decode(mime_doc.content), mime_doc.mime_type),
                    )
                )
            files = files_list

        url = f"/v1/deployments/extract/{project_id}/{iteration_id}"

        return PreparedRequest(method="POST", url=url, form_data=form_data, files=files)


class Deployments(SyncAPIResource, DeploymentsMixin):
    """Deployments API wrapper for managing deployment configurations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)

    def extract(
        self,
        project_id: str,
        iteration_id: str,
        document: Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl | None = None,
        documents: List[Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl] | None = None,
        temperature: float | None = None,
        seed: int | None = None,
        store: bool = True,
    ) -> RetabParsedChatCompletion:
        """Extract documents from a deployment.

        Args:
            project_id: ID of the project
            iteration_id: ID of the iteration
            document: Single document to process (mutually exclusive with documents)
            documents: List of documents to process (mutually exclusive with document)
            temperature: Optional temperature override
            seed: Optional seed for reproducibility
            store: Whether to store the results

        Returns:
            RetabParsedChatCompletion: The processing result
        """
        request = self.prepare_extract(project_id=project_id, iteration_id=iteration_id, document=document, documents=documents, temperature=temperature, seed=seed, store=store)
        response = self._client._prepared_request(request)
        return RetabParsedChatCompletion.model_validate(response)


class AsyncDeployments(AsyncAPIResource, DeploymentsMixin):
    """Async Deployments API wrapper for managing deployment configurations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)

    async def extract(
        self,
        project_id: str,
        iteration_id: str,
        document: Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl | None = None,
        documents: List[Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl] | None = None,
        temperature: float | None = None,
        seed: int | None = None,
        store: bool = True,
    ) -> RetabParsedChatCompletion:
        """Extract documents from a deployment.

        Args:
            project_id: ID of the project
            iteration_id: ID of the iteration
            document: Single document to process (mutually exclusive with documents)
            documents: List of documents to process (mutually exclusive with document)
            temperature: Optional temperature override
            seed: Optional seed for reproducibility
            store: Whether to store the results

        Returns:
            RetabParsedChatCompletion: The processing result
        """
        request = self.prepare_extract(project_id=project_id, iteration_id=iteration_id, document=document, documents=documents, temperature=temperature, seed=seed, store=store)
        response = await self._client._prepared_request(request)
        return RetabParsedChatCompletion.model_validate(response)



from typing import Any
import os

import nanoid  # type: ignore
import pytest

from retab import Retab


@pytest.mark.asyncio
async def test_mailboxes_crud(sync_client: Retab, company_json_schema: dict[str, Any], booking_confirmation_file_path_1: str) -> None:
    test_idx = nanoid.generate().lower()
    name = f"test_mailbox_{test_idx}"
    email_address = f"bert2_{test_idx}@{os.getenv('EMAIL_DOMAIN', 'mailbox.retab.com')}"
    webhook_url = "http://localhost:4000/product"
    model = "gpt-4o-mini"

    # First create a processor
    processor = sync_client.processors.create(
        name=name,
        json_schema=company_json_schema,
        model=model,
    )
    processor_id = processor.id

    try:
        # Create the mailbox
        mailbox = sync_client.processors.automations.mailboxes.create(
            processor_id=processor_id,
            name=name,
            email=email_address,
            webhook_url=webhook_url,
        )
        try:
            assert mailbox.email == email_address
            assert mailbox.webhook_url == webhook_url
            # Read
            mailbox = sync_client.processors.automations.mailboxes.get(email_address)
            assert mailbox.email == email_address
            # Update
            mailbox = sync_client.processors.automations.mailboxes.update(email_address, webhook_url="http://localhost:4000/product2")
            assert mailbox.webhook_url == "http://localhost:4000/product2"

            # TODO: send and email to email_address (need sendgrid account)

            # Delete the mailbox
            sync_client.processors.automations.mailboxes.delete(email_address)
            with pytest.raises(Exception):
                sync_client.processors.automations.mailboxes.get(email_address)
        finally:
            # Delete the mailbox if it was created
            try:
                sync_client.processors.automations.mailboxes.delete(email_address)
            except Exception:
                pass
    finally:
        # Delete the processor if it was created
        try:
            sync_client.processors.delete(processor_id)
        except Exception:
            pass



import datetime
from typing import Any, Optional

import nanoid  # type: ignore
from pydantic import BaseModel, Field, computed_field

from ...utils.json_schema import generate_schema_data_id, generate_schema_id
from ..inference_settings import InferenceSettings
from .documents import ProjectDocument
from .iterations import Iteration


class BaseProject(BaseModel):
    id: str = Field(default_factory=lambda: "proj_" + nanoid.generate())
    name: str = Field(default="", description="The name of the project")
    json_schema: dict[str, Any] = Field(default_factory=dict, description="The json schema of the project")
    default_inference_settings: InferenceSettings = Field(default=InferenceSettings(), description="The default inference properties for the project.")
    updated_at: datetime.datetime = Field(default_factory=lambda: datetime.datetime.now(tz=datetime.timezone.utc))


# Actual Object stored in DB
class Project(BaseProject):
    documents: list[ProjectDocument] = Field(default_factory=list)
    iterations: list[Iteration] = Field(default_factory=list)

    @computed_field  # type: ignore
    @property
    def schema_data_id(self) -> str:
        """Returns the SHA1 hash of the schema data, ignoring all prompt/description/default fields.

        Returns:
            str: A SHA1 hash string representing the schema data version.
        """
        return generate_schema_data_id(self.json_schema)

    # This is a computed field, it is exposed when serializing the object
    @computed_field  # type: ignore
    @property
    def schema_id(self) -> str:
        """Returns the SHA1 hash of the complete schema.

        Returns:
            str: A SHA1 hash string representing the complete schema version.
        """
        return generate_schema_id(self.json_schema)


class ListProjectParams(BaseModel):
    schema_id: Optional[str] = Field(default=None, description="The ID of the schema")
    schema_data_id: Optional[str] = Field(default=None, description="The ID of the schema data")


class CreateProjectRequest(BaseModel):
    name: str
    json_schema: dict[str, Any]
    default_inference_settings: InferenceSettings


# This is basically the same as BaseProject, but everything is optional.
# Could be achieved by convert_basemodel_to_partial_basemodel(BaseProject) but we prefer explicitness
class PatchProjectRequest(BaseModel):
    name: Optional[str] = Field(default=None, description="The name of the document")
    json_schema: Optional[dict[str, Any]] = Field(default=None, description="The json schema of the project")
    default_inference_settings: Optional[InferenceSettings] = Field(default=None, description="The default inference properties for the project (mostly used in the frontend)")


class AddIterationFromJsonlRequest(BaseModel):
    jsonl_gcs_path: str



from typing import Literal, Optional, Self

from pydantic import BaseModel, model_validator
from ..inference_settings import InferenceSettings

SelectionMode = Literal["all", "manual"]


# This is the input data for the prepare_dataset job
class PrepareDatasetInputData(BaseModel):
    dataset_id: Optional[str] = None
    schema_id: Optional[str] = None
    schema_data_id: Optional[str] = None

    selection_model: SelectionMode = "all"

    @model_validator(mode="after")
    def validate_input(self) -> Self:
        # The preference is:
        # 1. dataset_id
        # 2. schema_id
        # 3. schema_data_id
        if self.dataset_id is None and self.schema_id is None and self.schema_data_id is None:
            raise ValueError("At least one of dataset_id, schema_id, or schema_data_id must be provided")

        return self


# This is the input data for the split_dataset job
class DatasetSplitInputData(BaseModel):
    dataset_id: str
    train_size: Optional[int | float] = None
    eval_size: Optional[int | float] = None

    @model_validator(mode="after")
    def validate_input(self) -> Self:
        if self.train_size is not None and self.eval_size is not None:
            raise ValueError("train_size and eval_size cannot both be provided")
        return self


# This is the input data for the batch annotation job
class AnnotationInputData(BaseModel):
    data_file: str
    schema_id: str
    inference_settings: InferenceSettings


# This is the input data for the evaluation job
class ProjectInputData(BaseModel):
    eval_data_file: str
    schema_id: str
    inference_settings_1: InferenceSettings | None = None
    inference_settings_2: InferenceSettings



from typing import Optional

from pydantic import BaseModel

from ..inference_settings import InferenceSettings


class AnnotationInputData(BaseModel):
    dataset_id: str
    files_ids: Optional[list[str]] = None
    upsert: bool = False
    inference_settings: InferenceSettings



from typing import Any, Literal, Optional

from ....types.standards import FieldUnset

from ...._resource import AsyncAPIResource, SyncAPIResource
from ....utils.ai_models import assert_valid_model_extraction
from ....types.automations.endpoints import Endpoint, ListEndpoints, UpdateEndpointRequest
from ....types.standards import PreparedRequest


class EndpointsMixin:
    def prepare_create(
        self,
        processor_id: str,
        name: str,
        webhook_url: str,
        model: str = "gpt-4o-mini",
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
    ) -> PreparedRequest:
        assert_valid_model_extraction(model)

        endpoint_dict: dict[str, Any] = {
            'processor_id': processor_id,
            'name': name,
            'webhook_url': webhook_url,
        }
        if webhook_headers is not FieldUnset:
            endpoint_dict['webhook_headers'] = webhook_headers
        if need_validation is not FieldUnset:
            endpoint_dict['need_validation'] = need_validation

        request = Endpoint(**endpoint_dict)
        return PreparedRequest(method="POST", url="/v1/processors/automations/endpoints", data=request.model_dump(mode="json", exclude_unset=True))

    def prepare_list(
        self,
        processor_id: str,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        # Filtering parameters
        name: Optional[str] = None,
        webhook_url: Optional[str] = None,
    ) -> PreparedRequest:
        params = {
            "processor_id": processor_id,
            "before": before,
            "after": after,
            "limit": limit,
            "order": order,
            "name": name,
            "webhook_url": webhook_url,
        }
        # Remove None values
        params = {k: v for k, v in params.items() if v is not None}

        return PreparedRequest(method="GET", url="/v1/processors/automations/endpoints", params=params)

    def prepare_get(self, endpoint_id: str) -> PreparedRequest:
        """Get a specific endpoint configuration.

        Args:
            endpoint_id: ID of the endpoint

        Returns:
            Endpoint: The endpoint configuration
        """
        return PreparedRequest(method="GET", url=f"/v1/processors/automations/endpoints/{endpoint_id}")

    def prepare_update(
        self,
        endpoint_id: str,
        name: str = FieldUnset,
        default_language: str = FieldUnset,
        webhook_url: str = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
    ) -> PreparedRequest:
        update_dict: dict[str, Any] = {}
        if name is not FieldUnset:
            update_dict['name'] = name
        if default_language is not FieldUnset:
            update_dict['default_language'] = default_language
        if webhook_url is not FieldUnset:
            update_dict['webhook_url'] = webhook_url
        if webhook_headers is not FieldUnset:
            update_dict['webhook_headers'] = webhook_headers
        if need_validation is not FieldUnset:
            update_dict['need_validation'] = need_validation

        request = UpdateEndpointRequest(**update_dict)
        return PreparedRequest(method="PUT", url=f"/v1/processors/automations/endpoints/{endpoint_id}", data=request.model_dump(mode="json", exclude_unset=True))

    def prepare_delete(self, endpoint_id: str) -> PreparedRequest:
        return PreparedRequest(method="DELETE", url=f"/v1/processors/automations/endpoints/{endpoint_id}")


class Endpoints(SyncAPIResource, EndpointsMixin):
    """Endpoints API wrapper for managing endpoint configurations"""

    def create(
        self,
        processor_id: str,
        name: str,
        webhook_url: str,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
    ) -> Endpoint:
        """Create a new endpoint configuration.

        Args:
            name: Name of the endpoint
            webhook_url: Webhook endpoint URL
            json_schema: JSON schema for the endpoint
            webhook_headers: Optional HTTP headers for webhook requests
            image_resolution_dpi: Optional image resolution DPI
            browser_canvas: Optional browser canvas size
            modality: Processing modality (currently only "native" supported)
            model: AI model to use for processing
            temperature: Model temperature setting
            reasoning_effort: The effort level for the model to reason about the input data.
        Returns:
            Endpoint: The created endpoint configuration
        """
        request = self.prepare_create(
            processor_id=processor_id,
            name=name,
            webhook_url=webhook_url,
            webhook_headers=webhook_headers,
            need_validation=need_validation,
        )
        response = self._client._prepared_request(request)
        print(f"Endpoint Created. Url: https://www.retab.com/dashboard/processors/automations/{response['id']}")
        return Endpoint.model_validate(response)

    def list(
        self,
        processor_id: str,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        name: Optional[str] = None,
        webhook_url: Optional[str] = None,
    ) -> ListEndpoints:
        """List endpoint configurations with pagination support.

        Args:
            before: Optional cursor for pagination before a specific endpoint ID
            after: Optional cursor for pagination after a specific endpoint ID
            limit: Optional limit on number of results (max 100)
            order: Optional sort order ("asc" or "desc")
            name: Optional filter by endpoint name
            webhook_url: Optional filter by webhook URL

        Returns:
            ListEndpoints: Paginated list of endpoint configurations with metadata
        """
        request = self.prepare_list(processor_id, before, after, limit, order, name, webhook_url)
        response = self._client._prepared_request(request)
        return ListEndpoints.model_validate(response)

    def get(self, endpoint_id: str) -> Endpoint:
        """Get a specific endpoint configuration.

        Args:
            endpoint_id: ID of the endpoint

        Returns:
            Endpoint: The endpoint configuration
        """
        request = self.prepare_get(endpoint_id)
        response = self._client._prepared_request(request)
        return Endpoint.model_validate(response)

    def update(
        self,
        endpoint_id: str,
        name: str = FieldUnset,
        default_language: str = FieldUnset,
        webhook_url: str = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
    ) -> Endpoint:
        """Update an endpoint configuration.

        Args:
            id: ID of the endpoint to update
            name: New name for the endpoint
            webhook_url: New webhook URL
            webhook_headers: New webhook headers
            json_schema: New JSON schema for the endpoint
            image_resolution_dpi: New image resolution DPI
            browser_canvas: New browser canvas size
            modality: New processing modality
            model: New AI model
            temperature: New temperature setting
            reasoning_effort: The effort level for the model to reason about the input data.
        Returns:
            Endpoint: The updated endpoint configuration
        """
        request = self.prepare_update(
            endpoint_id=endpoint_id,
            name=name,
            default_language=default_language,
            webhook_url=webhook_url,
            webhook_headers=webhook_headers,
            need_validation=need_validation,
        )
        response = self._client._prepared_request(request)
        return Endpoint.model_validate(response)

    def delete(self, endpoint_id: str) -> None:
        """Delete an endpoint configuration.

        Args:
            id: ID of the endpoint to delete
        """
        request = self.prepare_delete(endpoint_id)
        self._client._prepared_request(request)
        print(f"Endpoint Deleted. ID: {endpoint_id}")


class AsyncEndpoints(AsyncAPIResource, EndpointsMixin):
    """Async Endpoints API wrapper for managing endpoint configurations"""

    async def create(
        self,
        processor_id: str,
        name: str,
        webhook_url: str,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
    ) -> Endpoint:
        request = self.prepare_create(
            processor_id=processor_id,
            name=name,
            webhook_url=webhook_url,
            webhook_headers=webhook_headers,
            need_validation=need_validation,
        )
        response = await self._client._prepared_request(request)
        print(f"Endpoint Created. Url: https://www.retab.com/dashboard/processors/automations/{response['id']}")

        return Endpoint.model_validate(response)

    async def list(
        self,
        processor_id: str,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        name: Optional[str] = None,
        webhook_url: Optional[str] = None,
    ) -> ListEndpoints:
        request = self.prepare_list(processor_id, before, after, limit, order, name, webhook_url)
        response = await self._client._prepared_request(request)
        return ListEndpoints.model_validate(response)

    async def get(self, endpoint_id: str) -> Endpoint:
        request = self.prepare_get(endpoint_id)
        response = await self._client._prepared_request(request)
        return Endpoint.model_validate(response)

    async def update(
        self,
        endpoint_id: str,
        name: str = FieldUnset,
        default_language: str = FieldUnset,
        webhook_url: str = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
    ) -> Endpoint:
        request = self.prepare_update(
            endpoint_id=endpoint_id,
            name=name,
            default_language=default_language,
            webhook_url=webhook_url,
            webhook_headers=webhook_headers,
            need_validation=need_validation,
        )
        response = await self._client._prepared_request(request)
        return Endpoint.model_validate(response)

    async def delete(self, endpoint_id: str) -> None:
        request = self.prepare_delete(endpoint_id)
        await self._client._prepared_request(request)



from typing import Any

import pytest

from retab import Retab


@pytest.mark.asyncio
async def test_preprocessor_crud(sync_client: Retab, company_json_schema: dict[str, Any]) -> None:
    name = "test_preprocessor"
    model = "gpt-4o-mini"

    # Create a processor
    processor = sync_client.processors.create(
        name=name,
        json_schema=company_json_schema,
        model=model,
    )
    processor_id = processor.id

    try:
        # Get the processor
        retrieved_processor = sync_client.processors.get(processor_id)
        assert retrieved_processor.name == name
        assert retrieved_processor.model == model

        # List processors
        processors = sync_client.processors.list(limit=10)
        assert len(processors.data) > 0
        assert any(p.id == processor_id for p in processors.data)

        # Update processor
        updated_name = "updated_test_preprocessor"
        updated_processor = sync_client.processors.update(
            processor_id,
            name=updated_name,
        )
        assert updated_processor.name == updated_name

        # Delete processor
        sync_client.processors.delete(processor_id)
        with pytest.raises(Exception):
            sync_client.processors.get(processor_id)
    finally:
        # Clean up
        try:
            sync_client.processors.delete(processor_id)
        except Exception:
            pass



from typing import Any, Literal, Optional
from pydantic import BaseModel, Field


class ReconciliationRequest(BaseModel):
    list_dicts: list[dict] = Field(description="List of dictionaries that will be reconciled into a single consensus dictionary.")
    reference_schema: Optional[dict[str, Any]] = Field(
        default=None,
        description="Optional schema defining the structure and types of the dictionary fields to validate the list of dictionaries against. Raise an error if one of the dictionaries does not match the schema.",
    )
    mode: Literal["direct", "aligned"] = Field(
        default="direct",
        description="The mode to use for the consensus. If 'direct', the consensus is computed directly from the list of dictionaries. If 'aligned', the consensus is computed from the aligned dictionaries.",
    )


class ReconciliationResponse(BaseModel):
    consensus_dict: dict = Field(description="The consensus dictionary containing the reconciled values from the input dictionaries.")
    likelihoods: dict = Field(description="A dictionary containing the likelihood/confidence scores for each field in the consensus dictionary.")



from typing import Any, Literal, Optional

from ...._resource import AsyncAPIResource, SyncAPIResource
from ....types.automations.links import Link, ListLinks, UpdateLinkRequest
from ....types.standards import PreparedRequest, FieldUnset


class LinksMixin:
    links_base_url: str = "/v1/processors/automations/links"

    def prepare_create(
        self,
        processor_id: str,
        name: str,
        webhook_url: str,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
        password: str | None = FieldUnset,
    ) -> PreparedRequest:
        link_dict: dict[str, Any] = {
            'processor_id': processor_id,
            'name': name,
            'webhook_url': webhook_url,
        }
        if webhook_headers is not FieldUnset:
            link_dict['webhook_headers'] = webhook_headers
        if need_validation is not FieldUnset:
            link_dict['need_validation'] = need_validation
        if password is not FieldUnset:
            link_dict['password'] = password

        request = Link(**link_dict)
        return PreparedRequest(method="POST", url=self.links_base_url, data=request.model_dump(mode="json", exclude_unset=True))

    def prepare_list(
        self,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        # Filtering parameters
        processor_id: Optional[str] = None,
        name: Optional[str] = None,
    ) -> PreparedRequest:
        params = {
            "before": before,
            "after": after,
            "limit": limit,
            "order": order,
            "processor_id": processor_id,
            "name": name,
        }
        # Remove None values
        params = {k: v for k, v in params.items() if v is not None}

        return PreparedRequest(method="GET", url=self.links_base_url, params=params)

    def prepare_get(self, link_id: str) -> PreparedRequest:
        """Get a specific extraction link configuration.

        Args:
            link_id: ID of the extraction link

        Returns:
            Link: The extraction link configuration
        """
        return PreparedRequest(method="GET", url=f"{self.links_base_url}/{link_id}")

    def prepare_update(
        self,
        link_id: str,
        name: str = FieldUnset,
        webhook_url: str = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
        password: str | None = FieldUnset,
    ) -> PreparedRequest:
        update_dict: dict[str, Any] = {}
        if name is not FieldUnset:
            update_dict['name'] = name
        if webhook_url is not FieldUnset:
            update_dict['webhook_url'] = webhook_url
        if webhook_headers is not FieldUnset:
            update_dict['webhook_headers'] = webhook_headers
        if need_validation is not FieldUnset:
            update_dict['need_validation'] = need_validation
        if password is not FieldUnset:
            update_dict['password'] = password

        request = UpdateLinkRequest(**update_dict)
        return PreparedRequest(method="PUT", url=f"{self.links_base_url}/{link_id}", data=request.model_dump(mode="json", exclude_unset=True, exclude_defaults=True))

    def prepare_delete(self, link_id: str) -> PreparedRequest:
        return PreparedRequest(method="DELETE", url=f"{self.links_base_url}/{link_id}", raise_for_status=True)


class Links(SyncAPIResource, LinksMixin):
    """Extraction Link API wrapper for managing extraction link configurations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)

    def create(
        self,
        processor_id: str,
        name: str,
        webhook_url: str,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
        password: str | None = FieldUnset,
    ) -> Link:
        """Create a new extraction link configuration.

        Args:
            name: Name of the extraction link
            json_schema: JSON schema to validate extracted data
            webhook_url: Webhook endpoint for forwarding processed files
            webhook_headers: Optional HTTP headers for webhook requests
            password: Optional password for protected links
            image_resolution_dpi: Optional image resolution DPI
            browser_canvas: Optional browser canvas
            modality: Processing modality (currently only "native" supported)
            model: AI model to use for processing
            temperature: Model temperature setting
            reasoning_effort: The effort level for the model to reason about the input data.
        Returns:
            Link: The created extraction link configuration
        """

        request = self.prepare_create(
            processor_id=processor_id,
            name=name,
            webhook_url=webhook_url,
            webhook_headers=webhook_headers,
            need_validation=need_validation,
            password=password,
        )
        response = self._client._prepared_request(request)

        print(f"Link Created. Url: https://www.retab.com/dashboard/processors/automations/{response['id']}")
        return Link.model_validate(response)

    def list(
        self,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        # Filtering parameters
        processor_id: Optional[str] = None,
        name: Optional[str] = None,
    ) -> ListLinks:
        """List extraction link configurations with pagination support.

        Args:
            before: Optional cursor for pagination before a specific link ID
            after: Optional cursor for pagination after a specific link ID
            limit: Optional limit on number of results (max 100)
            order: Optional sort order ("asc" or "desc")
            processor_id: Optional filter by processor ID
            name: Optional filter by link name

        Returns:
            ListLinks: Paginated list of extraction link configurations with metadata
        """
        request = self.prepare_list(before=before, after=after, limit=limit, order=order, processor_id=processor_id, name=name)
        response = self._client._prepared_request(request)
        return ListLinks.model_validate(response)

    def get(self, link_id: str) -> Link:
        """Get a specific extraction link configuration.

        Args:
            link_id: ID of the extraction link

        Returns:
            Link: The extraction link configuration
        """
        request = self.prepare_get(link_id)
        response = self._client._prepared_request(request)
        return Link.model_validate(response)

    def update(
        self,
        link_id: str,
        name: str = FieldUnset,
        webhook_url: str = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        password: str | None = FieldUnset,
        need_validation: bool = FieldUnset,
    ) -> Link:
        """Update an extraction link configuration.

        Args:
            link_id: ID of the extraction link to update
            name: New name for the link
            webhook_url: New webhook endpoint URL
            webhook_headers: New webhook headers
            password: New password for protected links
            image_resolution_dpi: New image resolution DPI
            browser_canvas: New browser canvas
            modality: New processing modality
            model: New AI model
            temperature: New temperature setting
            reasoning_effort: The effort level for the model to reason about the input data.
            json_schema: New JSON schema

        Returns:
            Link: The updated extraction link configuration
        """

        request = self.prepare_update(
            link_id=link_id,
            name=name,
            webhook_url=webhook_url,
            webhook_headers=webhook_headers,
            password=password,
            need_validation=need_validation,
        )
        response = self._client._prepared_request(request)
        return Link.model_validate(response)

    def delete(self, link_id: str) -> None:
        """Delete an extraction link configuration.

        Args:
            link_id: ID of the extraction link to delete

        Returns:
            Dict[str, str]: Response message confirming deletion
        """
        request = self.prepare_delete(link_id)
        self._client._prepared_request(request)


class AsyncLinks(AsyncAPIResource, LinksMixin):
    """Async Extraction Link API wrapper for managing extraction link configurations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)

    async def create(
        self,
        processor_id: str,
        name: str,
        webhook_url: str,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
        password: str | None = FieldUnset,
    ) -> Link:
        request = self.prepare_create(
            processor_id=processor_id,
            name=name,
            webhook_url=webhook_url,
            webhook_headers=webhook_headers,
            need_validation=need_validation,
            password=password,
        )
        response = await self._client._prepared_request(request)
        print(f"Link Created. Url: https://www.retab.com/dashboard/processors/automations/{response['id']}")
        return Link.model_validate(response)

    async def list(
        self,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        processor_id: Optional[str] = None,
        name: Optional[str] = None,
    ) -> ListLinks:
        request = self.prepare_list(before=before, after=after, limit=limit, order=order, processor_id=processor_id, name=name)
        response = await self._client._prepared_request(request)
        return ListLinks.model_validate(response)

    async def get(self, link_id: str) -> Link:
        request = self.prepare_get(link_id)
        response = await self._client._prepared_request(request)
        return Link.model_validate(response)

    async def update(
        self,
        link_id: str,
        name: str = FieldUnset,
        webhook_url: str = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        password: str | None = FieldUnset,
        need_validation: bool = FieldUnset,
    ) -> Link:
        request = self.prepare_update(
            link_id=link_id,
            name=name,
            webhook_url=webhook_url,
            webhook_headers=webhook_headers,
            password=password,
            need_validation=need_validation,
        )
        response = await self._client._prepared_request(request)
        return Link.model_validate(response)

    async def delete(self, link_id: str) -> None:
        request = self.prepare_delete(link_id)
        await self._client._prepared_request(request)



from typing import Any, Literal, List

from ....types.standards import FieldUnset

from ...._resource import AsyncAPIResource, SyncAPIResource
from ....types.automations.outlook import (
    FetchParams,
    ListOutlooks,
    MatchParams,
    Outlook,
    UpdateOutlookRequest,
)
from ....types.standards import PreparedRequest


class OutlooksMixin:
    outlooks_base_url: str = "/v1/processors/automations/outlook"

    def prepare_create(
        self,
        name: str,
        processor_id: str,
        webhook_url: str,
        default_language: str = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
        authorized_domains: list[str] = FieldUnset,
        authorized_emails: list[str] = FieldUnset,
        layout_schema: dict[str, Any] = FieldUnset,
        match_params: list[MatchParams] = FieldUnset,
        fetch_params: list[FetchParams] = FieldUnset,
    ) -> PreparedRequest:
        # Build outlook dictionary with only provided fields
        outlook_dict: dict[str, Any] = {
            'processor_id': processor_id,
            'name': name,
            'webhook_url': webhook_url,
        }
        if default_language is not FieldUnset:
            outlook_dict['default_language'] = default_language
        if webhook_headers is not FieldUnset:
            outlook_dict['webhook_headers'] = webhook_headers
        if need_validation is not FieldUnset:
            outlook_dict['need_validation'] = need_validation
        if authorized_domains is not FieldUnset:
            outlook_dict['authorized_domains'] = authorized_domains
        if authorized_emails is not FieldUnset:
            outlook_dict['authorized_emails'] = authorized_emails
        if layout_schema is not FieldUnset:
            outlook_dict['layout_schema'] = layout_schema
        if match_params is not FieldUnset:
            outlook_dict['match_params'] = match_params
        if fetch_params is not FieldUnset:
            outlook_dict['fetch_params'] = fetch_params

        # Validate the data
        outlook_data = Outlook(**outlook_dict)
        return PreparedRequest(
            method="POST",
            url=self.outlooks_base_url,
            data=outlook_data.model_dump(mode="json", exclude_unset=True),
        )

    def prepare_list(
        self,
        processor_id: str,
        before: str | None = None,
        after: str | None = None,
        limit: int = 10,
        order: Literal["asc", "desc"] | None = "desc",
        name: str | None = None,
        webhook_url: str | None = None,
    ) -> PreparedRequest:
        params = {
            "processor_id": processor_id,
            "before": before,
            "after": after,
            "limit": limit,
            "order": order,
            "name": name,
            "webhook_url": webhook_url,
        }
        # Remove None values
        params = {k: v for k, v in params.items() if v is not None}

        return PreparedRequest(method="GET", url=self.outlooks_base_url, params=params)

    def prepare_get(self, outlook_id: str) -> PreparedRequest:
        return PreparedRequest(method="GET", url=f"{self.outlooks_base_url}/{outlook_id}")

    def prepare_update(
        self,
        outlook_id: str,
        name: str = FieldUnset,
        default_language: str = FieldUnset,
        webhook_url: str = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
        authorized_domains: list[str] = FieldUnset,
        authorized_emails: list[str] = FieldUnset,
        match_params: list[MatchParams] = FieldUnset,
        fetch_params: list[FetchParams] = FieldUnset,
        layout_schema: dict[str, Any] = FieldUnset,
    ) -> PreparedRequest:
        update_dict: dict[str, Any] = {}
        if name is not FieldUnset:
            update_dict['name'] = name
        if default_language is not FieldUnset:
            update_dict['default_language'] = default_language
        if webhook_url is not FieldUnset:
            update_dict['webhook_url'] = webhook_url
        if webhook_headers is not FieldUnset:
            update_dict['webhook_headers'] = webhook_headers
        if need_validation is not FieldUnset:
            update_dict['need_validation'] = need_validation
        if authorized_domains is not FieldUnset:
            update_dict['authorized_domains'] = authorized_domains
        if authorized_emails is not FieldUnset:
            update_dict['authorized_emails'] = authorized_emails
        if layout_schema is not FieldUnset:
            update_dict['layout_schema'] = layout_schema
        if match_params is not FieldUnset:
            update_dict['match_params'] = match_params
        if fetch_params is not FieldUnset:
            update_dict['fetch_params'] = fetch_params

        update_outlook_request = UpdateOutlookRequest(**update_dict)

        return PreparedRequest(
            method="PUT",
            url=f"{self.outlooks_base_url}/{outlook_id}",
            data=update_outlook_request.model_dump(mode="json", exclude_unset=True),
        )

    def prepare_delete(self, outlook_id: str) -> PreparedRequest:
        return PreparedRequest(method="DELETE", url=f"{self.outlooks_base_url}/{outlook_id}")


class Outlooks(SyncAPIResource, OutlooksMixin):
    """Outlook API wrapper for managing outlook automation configurations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)

    def create(
        self,
        name: str,
        processor_id: str,
        webhook_url: str,
        default_language: str = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
        authorized_domains: list[str] = FieldUnset,
        authorized_emails: list[str] = FieldUnset,
        layout_schema: dict[str, Any] = FieldUnset,
        match_params: list[MatchParams] = FieldUnset,
        fetch_params: list[FetchParams] = FieldUnset,
    ) -> Outlook:
        """Create a new outlook automation configuration.

        Args:
            name: Name of the outlook plugin
            processor_id: ID of the processor to use for the automation
            webhook_url: Webhook URL to receive processed data
            webhook_headers: Webhook headers to send with processed data
            authorized_domains: List of authorized domains
            authorized_emails: List of authorized emails
            layout_schema: Layout schema to display the data
            match_params: List of match parameters for the outlook automation
            fetch_params: List of fetch parameters for the outlook automation
        Returns:
            Outlook: The created outlook plugin configuration
        """

        request = self.prepare_create(
            name=name,
            processor_id=processor_id,
            webhook_url=webhook_url,
            default_language=default_language,
            webhook_headers=webhook_headers,
            need_validation=need_validation,
            authorized_domains=authorized_domains,
            authorized_emails=authorized_emails,
            layout_schema=layout_schema,
            match_params=match_params,
            fetch_params=fetch_params,
        )
        response = self._client._prepared_request(request)

        print(f"Outlook plugin created. Url: https://www.retab.com/dashboard/processors/automations/{response['id']}")

        return Outlook.model_validate(response)

    def list(
        self,
        processor_id: str,
        before: str | None = None,
        after: str | None = None,
        limit: int = 10,
        order: Literal["asc", "desc"] | None = "desc",
        name: str | None = None,
        webhook_url: str | None = None,
    ) -> ListOutlooks:
        """List all outlook automation configurations.

        Args:
            before: Optional cursor for pagination - get results before this log ID
            after: Optional cursor for pagination - get results after this log ID
            limit: Maximum number of logs to return (1-100, default 10)
            order: Sort order by creation time - "asc" or "desc" (default "desc")
            name: Optional name filter
            webhook_url: Optional webhook URL filter
        Returns:
            List[Outlook]: List of outlook plugin configurations
        """
        request = self.prepare_list(processor_id, before, after, limit, order, name, webhook_url)
        response = self._client._prepared_request(request)
        return ListOutlooks.model_validate(response)

    def get(self, outlook_id: str) -> Outlook:
        """Get a specific outlook automation configuration.

        Args:
            id: ID of the outlook plugin

        Returns:
            Outlook: The outlook plugin configuration
        """
        request = self.prepare_get(outlook_id)
        response = self._client._prepared_request(request)
        return Outlook.model_validate(response)

    def update(
        self,
        outlook_id: str,
        name: str = FieldUnset,
        default_language: str = FieldUnset,
        webhook_url: str = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
        authorized_domains: List[str] = FieldUnset,
        authorized_emails: List[str] = FieldUnset,
        layout_schema: dict[str, Any] = FieldUnset,
        match_params: List[MatchParams] = FieldUnset,
        fetch_params: List[FetchParams] = FieldUnset,
    ) -> Outlook:
        """Update an outlook automation configuration.

        Args:
            outlook_id: ID of the outlook plugin to update
            name: New name for the outlook plugin
            webhook_url: New webhook URL
            webhook_headers: New webhook headers
            authorized_domains: New authorized domains
            authorized_emails: New authorized emails
            match_params: New match parameters for the outlook automation
            fetch_params: New fetch parameters for the outlook automation
            layout_schema: New layout schema for the outlook automation

        Returns:
            Outlook: The updated outlook plugin configuration
        """
        request = self.prepare_update(
            outlook_id,
            name=name,
            default_language=default_language,
            webhook_url=webhook_url,
            webhook_headers=webhook_headers,
            need_validation=need_validation,
            authorized_domains=authorized_domains,
            authorized_emails=authorized_emails,
            layout_schema=layout_schema,
            match_params=match_params,
            fetch_params=fetch_params,
        )
        response = self._client._prepared_request(request)
        return Outlook.model_validate(response)

    def delete(self, outlook_id: str) -> None:
        """Delete an outlook automation configuration.

        Args:
            outlook_id: ID of the outlook plugin to delete
        """
        request = self.prepare_delete(outlook_id)
        self._client._prepared_request(request)
        return None


class AsyncOutlooks(AsyncAPIResource, OutlooksMixin):
    def __init__(self, client: Any) -> None:
        super().__init__(client=client)

    async def create(
        self,
        name: str,
        processor_id: str,
        webhook_url: str,
        default_language: str = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
        authorized_domains: list[str] = FieldUnset,
        authorized_emails: list[str] = FieldUnset,
        layout_schema: dict[str, Any] = FieldUnset,
        match_params: list[MatchParams] = FieldUnset,
        fetch_params: list[FetchParams] = FieldUnset,
    ) -> Outlook:
        request = self.prepare_create(
            name=name,
            processor_id=processor_id,
            webhook_url=webhook_url,
            default_language=default_language,
            webhook_headers=webhook_headers,
            need_validation=need_validation,
            authorized_domains=authorized_domains,
            authorized_emails=authorized_emails,
            layout_schema=layout_schema,
            match_params=match_params,
            fetch_params=fetch_params,
        )
        response = await self._client._prepared_request(request)
        print(f"Outlook plugin created. Url: https://www.retab.com/dashboard/processors/automations/{response['id']}")
        return Outlook.model_validate(response)

    async def list(
        self,
        processor_id: str,
        before: str | None = None,
        after: str | None = None,
        limit: int = 10,
        order: Literal["asc", "desc"] | None = "desc",
        name: str | None = None,
        webhook_url: str | None = None,
    ) -> ListOutlooks:
        request = self.prepare_list(processor_id, before, after, limit, order, name, webhook_url)
        response = await self._client._prepared_request(request)
        return ListOutlooks.model_validate(response)

    async def get(self, outlook_id: str) -> Outlook:
        request = self.prepare_get(outlook_id)
        response = await self._client._prepared_request(request)
        return Outlook.model_validate(response)

    async def update(
        self,
        outlook_id: str,
        name: str = FieldUnset,
        default_language: str = FieldUnset,
        webhook_url: str = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
        authorized_domains: List[str] = FieldUnset,
        authorized_emails: List[str] = FieldUnset,
        layout_schema: dict[str, Any] = FieldUnset,
        match_params: List[MatchParams] = FieldUnset,
        fetch_params: List[FetchParams] = FieldUnset,
    ) -> Outlook:
        request = self.prepare_update(
            outlook_id=outlook_id,
            name=name,
            default_language=default_language,
            webhook_url=webhook_url,
            webhook_headers=webhook_headers,
            need_validation=need_validation,
            authorized_domains=authorized_domains,
            authorized_emails=authorized_emails,
            layout_schema=layout_schema,
            match_params=match_params,
            fetch_params=fetch_params,
        )
        response = await self._client._prepared_request(request)
        return Outlook.model_validate(response)

    async def delete(self, outlook_id: str) -> None:
        request = self.prepare_delete(outlook_id)
        await self._client._prepared_request(request)
        return None



import nanoid  # type: ignore
from pydantic import BaseModel, Field, computed_field

from ..logs import AutomationConfig, UpdateAutomationRequest
from ..pagination import ListMetadata


class Endpoint(AutomationConfig):
    @computed_field
    @property
    def object(self) -> str:
        return "automation.endpoint"

    id: str = Field(default_factory=lambda: "endp_" + nanoid.generate(), description="Unique identifier for the extraction endpoint")


class ListEndpoints(BaseModel):
    data: list[Endpoint]
    list_metadata: ListMetadata


# Inherits from the methods of UpdateAutomationRequest
class UpdateEndpointRequest(UpdateAutomationRequest):
    pass



import re
from typing import Any, Dict, List, Optional

import nanoid  # type: ignore
from pydantic import BaseModel, EmailStr, Field, computed_field, field_validator

from ..logs import AutomationConfig, UpdateAutomationRequest
from ..pagination import ListMetadata

domain_pattern = re.compile(r"^(?:[a-zA-Z0-9-]+\.)+[a-zA-Z]{2,}$")


class AutomationLevel(BaseModel):
    distance_threshold: float = Field(default=0.9, description="Distance threshold for the automation")
    score_threshold: float = Field(default=0.9, description="Score threshold for the automation")


class MatchParams(BaseModel):
    endpoint: str = Field(..., description="Endpoint for matching parameters")
    headers: Dict[str, str] = Field(..., description="Headers for the request")
    path: str = Field(..., description="Path for matching parameters")


class FetchParams(BaseModel):
    endpoint: str = Field(..., description="Endpoint for fetching parameters")
    headers: Dict[str, str] = Field(..., description="Headers for the request")
    name: str = Field(..., description="Name of the fetch parameter")


class Outlook(AutomationConfig):
    @computed_field
    @property
    def object(self) -> str:
        return "automation.outlook"

    id: str = Field(default_factory=lambda: "outlook_" + nanoid.generate(), description="Unique identifier for the outlook")

    authorized_domains: list[str] = Field(default_factory=list, description="List of authorized domains to receive the emails from")
    authorized_emails: List[EmailStr] = Field(default_factory=list, description="List of emails to access the link")

    layout_schema: Optional[dict[str, Any]] = Field(default=None, description="Layout schema format used to display the data")

    # Optional Fields for data integration
    match_params: List[MatchParams] = Field(default_factory=list, description="List of match parameters for the outlook automation")
    fetch_params: List[FetchParams] = Field(default_factory=list, description="List of fetch parameters for the outlook automation")

    # @model_validator(mode="before")
    # @classmethod
    # def compute_layout_schema(cls, values: dict[str, Any]) -> dict[str, Any]:
    #     if values.get("layout_schema") is None:
    #         values["layout_schema"] = convert_schema_to_layout(values["json_schema"])
    #     return values


class ListOutlooks(BaseModel):
    data: list[Outlook]
    list_metadata: ListMetadata


# Inherits from the methods of UpdateAutomationRequest
class UpdateOutlookRequest(UpdateAutomationRequest):
    authorized_domains: Optional[list[str]] = None
    authorized_emails: Optional[List[EmailStr]] = None

    match_params: Optional[List[MatchParams]] = None
    fetch_params: Optional[List[FetchParams]] = None

    layout_schema: Optional[dict[str, Any]] = None

    @field_validator("authorized_emails", mode="before")
    def normalize_authorized_emails(cls, emails: Optional[List[str]]) -> Optional[List[str]]:
        return [email.strip().lower() for email in emails] if emails else None



from io import IOBase
from pathlib import Path
from typing import Any, List, Literal, Optional

from pydantic import EmailStr, HttpUrl
from ....types.standards import FieldUnset

from ...._resource import AsyncAPIResource, SyncAPIResource
from ....utils.mime import prepare_mime_document
from ....types.automations.mailboxes import ListMailboxes, Mailbox, UpdateMailboxRequest
from ....types.mime import EmailData, MIMEData
from ....types.standards import PreparedRequest


class MailBoxesMixin:
    mailboxes_base_url: str = "/v1/processors/automations/mailboxes"

    def prepare_create(
        self,
        email: str,
        name: str,
        processor_id: str,
        webhook_url: str,
        authorized_domains: list[str] = FieldUnset,
        authorized_emails: list[EmailStr] = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        default_language: str = FieldUnset,
        need_validation: bool = FieldUnset,
    ) -> PreparedRequest:
        """Create a new email automation configuration."""
        mailbox_dict: dict[str, Any] = {
            'name': name,
            'processor_id': processor_id,
            'webhook_url': webhook_url,
            'email': email,
        }
        if default_language is not FieldUnset:
            mailbox_dict['default_language'] = default_language
        if webhook_headers is not FieldUnset:
            mailbox_dict['webhook_headers'] = webhook_headers
        if need_validation is not FieldUnset:
            mailbox_dict['need_validation'] = need_validation
        if authorized_domains is not FieldUnset:
            mailbox_dict['authorized_domains'] = authorized_domains
        if authorized_emails is not FieldUnset:
            mailbox_dict['authorized_emails'] = authorized_emails

        mailbox = Mailbox(**mailbox_dict)
        return PreparedRequest(method="POST", url=self.mailboxes_base_url, data=mailbox.model_dump(mode="json", exclude_unset=True))

    def prepare_list(
        self,
        processor_id: str,
        before: str | None = None,
        after: str | None = None,
        limit: int = 10,
        order: Literal["asc", "desc"] | None = "desc",
        email: Optional[str] = None,
        name: Optional[str] = None,
        webhook_url: Optional[str] = None,
    ) -> PreparedRequest:
        params = {
            "processor_id": processor_id,
            "before": before,
            "after": after,
            "limit": limit,
            "order": order,
            "email": email,
            "name": name,
            "webhook_url": webhook_url,
        }
        # Remove None values
        params = {k: v for k, v in params.items() if v is not None}

        return PreparedRequest(method="GET", url=self.mailboxes_base_url, params=params)

    def prepare_get(self, mailbox_id: str) -> PreparedRequest:
        return PreparedRequest(method="GET", url=f"{self.mailboxes_base_url}/{mailbox_id}")

    def prepare_update(
        self,
        mailbox_id: str,
        name: str = FieldUnset,
        default_language: str = FieldUnset,
        webhook_url: str = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
        authorized_domains: list[str] = FieldUnset,
        authorized_emails: list[str] = FieldUnset,
    ) -> PreparedRequest:
        update_dict: dict[str, Any] = {}
        if name is not FieldUnset:
            update_dict['name'] = name
        if default_language is not FieldUnset:
            update_dict['default_language'] = default_language
        if webhook_url is not FieldUnset:
            update_dict['webhook_url'] = webhook_url
        if webhook_headers is not FieldUnset:
            update_dict['webhook_headers'] = webhook_headers
        if need_validation is not FieldUnset:
            update_dict['need_validation'] = need_validation
        if authorized_domains is not FieldUnset:
            update_dict['authorized_domains'] = authorized_domains
        if authorized_emails is not FieldUnset:
            update_dict['authorized_emails'] = authorized_emails

        update_mailbox_request = UpdateMailboxRequest(**update_dict)
        return PreparedRequest(method="PUT", url=f"/v1/processors/automations/mailboxes/{mailbox_id}", data=update_mailbox_request.model_dump(mode="json", exclude_unset=True))

    def prepare_delete(self, mailbox_id: str) -> PreparedRequest:
        return PreparedRequest(method="DELETE", url=f"/v1/processors/automations/mailboxes/{mailbox_id}", raise_for_status=True)


class Mailboxes(SyncAPIResource, MailBoxesMixin):
    """Emails API wrapper for managing email automation configurations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.tests = TestMailboxes(client=client)

    def create(
        self,
        email: str,
        name: str,
        webhook_url: str,
        processor_id: str,
        authorized_domains: list[str] = FieldUnset,
        authorized_emails: list[EmailStr] = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        default_language: str = FieldUnset,
        need_validation: bool = FieldUnset,
    ) -> Mailbox:
        """Create a new email automation configuration.

        Args:
            email: Email address for the mailbox
            name: Name of the mailbox
            webhook_url: Webhook URL to receive processed emails
            processor_id: ID of the processor to use for the mailbox
            authorized_domains: List of authorized domains for the mailbox
            authorized_emails: List of authorized emails for the mailbox
            webhook_headers: Webhook headers to send with processed emails

        Returns:
            Mailbox: The created mailbox configuration
        """

        request = self.prepare_create(
            email=email,
            name=name,
            processor_id=processor_id,
            webhook_url=webhook_url,
            authorized_domains=authorized_domains,
            authorized_emails=authorized_emails,
            webhook_headers=webhook_headers,
            default_language=default_language,
            need_validation=need_validation,
        )
        response = self._client._prepared_request(request)

        print(f"Mailbox {response['email']} created. Url: https://www.retab.com/dashboard/processors/automations/{response['id']}")

        return Mailbox.model_validate(response)

    def list(
        self,
        processor_id: str,
        before: str | None = None,
        after: str | None = None,
        limit: int = 10,
        order: Literal["asc", "desc"] | None = "desc",
        name: Optional[str] = None,
        email: Optional[str] = None,
        webhook_url: Optional[str] = None,
    ) -> ListMailboxes:
        """List all email automation configurations.

        Args:
            before: Optional cursor for pagination - get results before this log ID
            after: Optional cursor for pagination - get results after this log ID
            limit: Maximum number of logs to return (1-100, default 10)
            order: Sort order by creation time - "asc" or "desc" (default "desc")
            email: Optional email address filter
            webhook_url: Optional webhook URL filter
            schema_id: Optional schema ID filter
            schema_data_id: Optional schema data ID filter

        Returns:
            ListMailboxes: List of mailbox configurations
        """
        request = self.prepare_list(
            processor_id=processor_id,
            before=before,
            after=after,
            limit=limit,
            order=order,
            email=email,
            name=name,
            webhook_url=webhook_url,
        )
        response = self._client._prepared_request(request)
        return ListMailboxes.model_validate(response)

    def get(self, mailbox_id: str) -> Mailbox:
        """Get a specific email automation configuration.

        Args:
            mailbox_id: ID of the mailbox

        Returns:
            Mailbox: The mailbox configuration
        """
        request = self.prepare_get(mailbox_id)
        response = self._client._prepared_request(request)
        return Mailbox.model_validate(response)

    def update(
        self,
        mailbox_id: str,
        name: str = FieldUnset,
        default_language: str = FieldUnset,
        webhook_url: str = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
        authorized_domains: List[str] = FieldUnset,
        authorized_emails: List[str] = FieldUnset,
    ) -> Mailbox:
        """Update an email automation configuration.

        Args:
            email: Email address of the mailbox to update
            name: New name
            default_language: New default language
            webhook_url: New webhook configuration
            webhook_headers: New webhook configuration
            need_validation: New webhook configuration
            authorized_domains: New webhook configuration
            authorized_emails: New webhook configuration

        Returns:
            Mailbox: The updated mailbox configuration
        """
        request = self.prepare_update(
            mailbox_id=mailbox_id,
            name=name,
            default_language=default_language,
            webhook_url=webhook_url,
            webhook_headers=webhook_headers,
            need_validation=need_validation,
            authorized_domains=authorized_domains,
            authorized_emails=authorized_emails,
        )
        response = self._client._prepared_request(request)
        return Mailbox.model_validate(response)

    def delete(self, mailbox_id: str) -> None:
        """Delete an email automation configuration.

        Args:
            email: Email address of the mailbox to delete
        """
        request = self.prepare_delete(mailbox_id)
        self._client._prepared_request(request)
        return None


class AsyncMailboxes(AsyncAPIResource, MailBoxesMixin):
    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.tests = AsyncTestMailboxes(client=client)

    async def create(
        self,
        email: str,
        name: str,
        webhook_url: str,
        processor_id: str,
        authorized_domains: List[str] = FieldUnset,
        authorized_emails: List[EmailStr] = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        default_language: str = FieldUnset,
        need_validation: bool = FieldUnset,
    ) -> Mailbox:
        request = self.prepare_create(
            email=email,
            name=name,
            processor_id=processor_id,
            webhook_url=webhook_url,
            authorized_domains=authorized_domains,
            authorized_emails=authorized_emails,
            webhook_headers=webhook_headers,
            default_language=default_language,
            need_validation=need_validation,
        )
        response = await self._client._prepared_request(request)

        print(f"Mailbox {response['email']} created. Url: https://www.retab.com/dashboard/processors/automations/{response['id']}")

        return Mailbox.model_validate(response)

    async def list(
        self,
        processor_id: str,
        before: str | None = None,
        after: str | None = None,
        limit: int = 10,
        order: Literal["asc", "desc"] | None = "desc",
        email: str | None = None,
        name: str | None = None,
        webhook_url: str | None = None,
    ) -> ListMailboxes:
        request = self.prepare_list(
            processor_id=processor_id,
            before=before,
            after=after,
            limit=limit,
            order=order,
            email=email,
            name=name,
            webhook_url=webhook_url,
        )
        response = await self._client._prepared_request(request)
        return ListMailboxes.model_validate(response)

    async def get(self, mailbox_id: str) -> Mailbox:
        request = self.prepare_get(mailbox_id)
        response = await self._client._prepared_request(request)
        return Mailbox.model_validate(response)

    async def update(
        self,
        mailbox_id: str,
        name: str = FieldUnset,
        default_language: str = FieldUnset,
        webhook_url: str = FieldUnset,
        webhook_headers: dict[str, str] = FieldUnset,
        need_validation: bool = FieldUnset,
        authorized_domains: List[str] = FieldUnset,
        authorized_emails: List[str] = FieldUnset,
    ) -> Mailbox:
        request = self.prepare_update(
            mailbox_id=mailbox_id,
            name=name,
            default_language=default_language,
            webhook_url=webhook_url,
            webhook_headers=webhook_headers,
            need_validation=need_validation,
            authorized_domains=authorized_domains,
            authorized_emails=authorized_emails,
        )
        response = await self._client._prepared_request(request)
        return Mailbox.model_validate(response)

    async def delete(self, mailbox_id: str) -> None:
        request = self.prepare_delete(mailbox_id)
        await self._client._prepared_request(request)
        return None


class TestMailboxesMixin:
    def prepare_forward(
        self,
        mailbox_id: str,
        document: Path | str | IOBase | HttpUrl | MIMEData,
        verbose: bool = True,
    ) -> PreparedRequest:
        mime_document = prepare_mime_document(document)
        return PreparedRequest(method="POST", url=f"/v1/processors/automations/mailboxes/tests/forward/{mailbox_id}", data={"document": mime_document.model_dump()})

    def print_forward_verbose(self, email_data: EmailData) -> None:
        print("\nTEST EMAIL FORWARDING RESULTS:")
        print("\n#########################")
        print(f"Email ID: {email_data.id}")
        print(f"Subject: {email_data.subject}")
        print(f"From: {email_data.sender}")
        print(f"To: {', '.join(str(r) for r in email_data.recipients_to)}")
        if email_data.recipients_cc:
            print(f"CC: {', '.join(str(r) for r in email_data.recipients_cc)}")
        print(f"Sent at: {email_data.sent_at}")
        print(f"Attachments: {len(email_data.attachments)}")
        if email_data.body_plain:
            print("\nBody Preview:")
            print(email_data.body_plain[:500] + "..." if len(email_data.body_plain) > 500 else email_data.body_plain)


class TestMailboxes(SyncAPIResource, TestMailboxesMixin):
    def forward(
        self,
        mailbox_id: str,
        document: Path | str | IOBase | HttpUrl | MIMEData,
        verbose: bool = True,
    ) -> EmailData:
        """Mock endpoint that simulates the complete email forwarding process with sample data.

        Args:
            email: Email address of the mailbox to mock

        Returns:
            DocumentExtractResponse: The simulated extraction response
        """
        request = self.prepare_forward(mailbox_id, document, verbose)
        response = self._client._prepared_request(request)

        email_data = EmailData.model_validate(response)

        if verbose:
            self.print_forward_verbose(email_data)
        return email_data


class AsyncTestMailboxes(AsyncAPIResource, TestMailboxesMixin):
    async def forward(
        self,
        mailbox_id: str,
        document: Path | str | IOBase | HttpUrl | MIMEData,
        verbose: bool = True,
    ) -> EmailData:
        request = self.prepare_forward(mailbox_id, document, verbose)
        response = await self._client._prepared_request(request)
        email_data = EmailData.model_validate(response)
        if verbose:
            self.print_forward_verbose(email_data)
        return email_data



import base64
import json
from io import IOBase
from pathlib import Path
from typing import Any, AsyncGenerator, Generator

from anthropic.types.message_param import MessageParam
from openai.types.chat import ChatCompletionMessageParam
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from openai.types.chat.parsed_chat_completion import ParsedChatCompletionMessage
from openai.types.responses.response import Response
from openai.types.responses.response_input_param import ResponseInputItemParam
from pydantic import HttpUrl

from ..._resource import AsyncAPIResource, SyncAPIResource
from ...utils.ai_models import assert_valid_model_extraction
from ...utils.json_schema import filter_auxiliary_fields_json, load_json_schema, unflatten_dict
from ...utils.mime import MIMEData, prepare_mime_document
from ...utils.stream_context_managers import as_async_context_manager, as_context_manager
from ...types.chat import ChatCompletionRetabMessage
from ...types.documents.extractions import DocumentExtractRequest, LogExtractionRequest, RetabParsedChatCompletion, RetabParsedChatCompletionChunk, RetabParsedChoice
from ...types.browser_canvas import BrowserCanvas
from ...types.modalities import Modality
from ...types.schemas.object import Schema
from ...types.standards import PreparedRequest, FieldUnset


def maybe_parse_to_pydantic(schema: Schema, response: RetabParsedChatCompletion, allow_partial: bool = False) -> RetabParsedChatCompletion:
    if response.choices[0].message.content:
        try:
            if allow_partial:
                response.choices[0].message.parsed = schema._partial_pydantic_model.model_validate(filter_auxiliary_fields_json(response.choices[0].message.content))
            else:
                response.choices[0].message.parsed = schema.pydantic_model.model_validate(filter_auxiliary_fields_json(response.choices[0].message.content))
        except Exception:
            pass
    return response


class BaseExtractionsMixin:
    def prepare_extraction(
        self,
        json_schema: dict[str, Any] | Path | str,
        document: Path | str | IOBase | HttpUrl | None = None,
        documents: list[Path | str | IOBase | HttpUrl] | None = None,
        image_resolution_dpi: int = FieldUnset,
        browser_canvas: BrowserCanvas = FieldUnset,
        model: str = FieldUnset,
        temperature: float = FieldUnset,
        modality: Modality = FieldUnset,
        reasoning_effort: ChatCompletionReasoningEffort = FieldUnset,
        stream: bool = False,
        n_consensus: int = FieldUnset,
        store: bool = False,
        idempotency_key: str | None = None,
    ) -> PreparedRequest:
        assert_valid_model_extraction(model)

        json_schema = load_json_schema(json_schema)

        # Handle both single document and multiple documents
        if document is not None and documents is not None:
            raise ValueError("Cannot provide both 'document' and 'documents' parameters. Use either one.")

        # Convert single document to documents list for consistency
        if document is not None:
            processed_documents = [prepare_mime_document(document)]
        elif documents is not None:
            processed_documents = [prepare_mime_document(doc) for doc in documents]
        else:
            raise ValueError("Must provide either 'document' or 'documents' parameter.")

        # Build request dictionary with only provided fields
        request_dict = {
            'json_schema': json_schema,
            'documents': processed_documents,
            'stream': stream,
            'store': store,
        }
        if model is not FieldUnset:
            request_dict['model'] = model
        if temperature is not FieldUnset:
            request_dict['temperature'] = temperature
        if modality is not FieldUnset:
            request_dict['modality'] = modality
        if reasoning_effort is not FieldUnset:
            request_dict['reasoning_effort'] = reasoning_effort
        if n_consensus is not FieldUnset:
            request_dict['n_consensus'] = n_consensus
        if image_resolution_dpi is not FieldUnset:
            request_dict['image_resolution_dpi'] = image_resolution_dpi
        if browser_canvas is not FieldUnset:
            request_dict['browser_canvas'] = browser_canvas

        # Validate DocumentAPIRequest data (raises exception if invalid)
        request = DocumentExtractRequest(**request_dict)

        return PreparedRequest(
            method="POST", url="/v1/documents/extractions", data=request.model_dump(mode="json", exclude_unset=True, exclude_defaults=True), idempotency_key=idempotency_key
        )

    def prepare_log_extraction(
        self,
        document: Path | str | IOBase | HttpUrl | None,
        json_schema: dict[str, Any],
        model: str,
        temperature: float,
        completion: Any | None = None,
        # The messages can be provided in different formats, we will convert them to the Retab-compatible format
        messages: list[ChatCompletionRetabMessage] | None = None,
        openai_messages: list[ChatCompletionMessageParam] | None = None,
        anthropic_messages: list[MessageParam] | None = None,
        anthropic_system_prompt: str | None = None,
        # New fields for the Responses API
        openai_responses_input: list[ResponseInputItemParam] | None = None,
        openai_responses_output: Response | None = None,
    ) -> PreparedRequest:
        if document is None:
            mime_document = MIMEData(
                filename="dummy.txt",
                # url is a base64 encoded string with the mime type and the content. For the dummy one we will send a .txt file with the text "No document provided"
                url="data:text/plain;base64," + base64.b64encode(b"No document provided").decode("utf-8"),
            )
        else:
            mime_document = prepare_mime_document(document)

        return PreparedRequest(
            method="POST",
            url="/v1/documents/log_extraction",
            data=LogExtractionRequest(
                document=mime_document,
                messages=messages,
                openai_messages=openai_messages,
                anthropic_messages=anthropic_messages,
                anthropic_system_prompt=anthropic_system_prompt,
                completion=completion,
                openai_responses_input=openai_responses_input,
                openai_responses_output=openai_responses_output,
                json_schema=json_schema,
                model=model,
                temperature=temperature,
            ).model_dump(mode="json"),
            raise_for_status=True,
        )


class Extractions(SyncAPIResource, BaseExtractionsMixin):
    """Extraction API wrapper"""

    def parse(
        self,
        json_schema: dict[str, Any] | Path | str,
        model: str,
        document: Path | str | IOBase | HttpUrl | None = None,
        documents: list[Path | str | IOBase | HttpUrl] | None = None,
        image_resolution_dpi: int = FieldUnset,
        browser_canvas: BrowserCanvas = FieldUnset,
        temperature: float = FieldUnset,
        modality: Modality = FieldUnset,
        reasoning_effort: ChatCompletionReasoningEffort = FieldUnset,
        n_consensus: int = FieldUnset,
        idempotency_key: str | None = None,
        store: bool = False,
    ) -> RetabParsedChatCompletion:
        """
        Process one or more documents using the Retab API.

        Args:
            json_schema: JSON schema defining the expected data structure
            model: The AI model to use for processing
            document: Single document to process (use either this or documents, not both)
            documents: List of documents to process (use either this or document, not both)
            image_resolution_dpi: Optional image resolution DPI
            browser_canvas: Optional browser canvas size
            temperature: Model temperature setting (0-1)
            modality: Modality of the document (e.g., native)
            reasoning_effort: The effort level for the model to reason about the input data.
            n_consensus: Number of consensus extractions to perform (default: 1 which computes a single extraction and the likelihoods comes from the model logprobs)
            idempotency_key: Idempotency key for request
            store: Whether to store the document in the Retab database
        Returns:
            RetabParsedChatCompletion: Parsed response from the API
        Raises:
            ValueError: If neither document nor documents is provided, or if both are provided
            HTTPException: If the request fails
        """

        # Validate DocumentAPIRequest data (raises exception if invalid)
        request = self.prepare_extraction(
            json_schema=json_schema,
            document=document,
            documents=documents,
            image_resolution_dpi=image_resolution_dpi,
            browser_canvas=browser_canvas,
            model=model,
            temperature=temperature,
            modality=modality,
            reasoning_effort=reasoning_effort,
            stream=False,
            n_consensus=n_consensus,
            store=store,
            idempotency_key=idempotency_key,
        )
        response = self._client._prepared_request(request)

        schema = Schema(json_schema=load_json_schema(json_schema))
        return maybe_parse_to_pydantic(schema, RetabParsedChatCompletion.model_validate(response))

    @as_context_manager
    def stream(
        self,
        json_schema: dict[str, Any] | Path | str,
        model: str,
        document: Path | str | IOBase | HttpUrl | None = None,
        documents: list[Path | str | IOBase | HttpUrl] | None = None,
        image_resolution_dpi: int = FieldUnset,
        browser_canvas: BrowserCanvas = FieldUnset,
        temperature: float = FieldUnset,
        modality: Modality = FieldUnset,
        reasoning_effort: ChatCompletionReasoningEffort = FieldUnset,
        n_consensus: int = FieldUnset,
        idempotency_key: str | None = None,
        store: bool = False,
    ) -> Generator[RetabParsedChatCompletion, None, None]:
        """
        Process one or more documents using the Retab API with streaming enabled.

        Args:
            json_schema: JSON schema defining the expected data structure
            model: The AI model to use for processing
            document: Single document to process (use either this or documents, not both)
            documents: List of documents to process (use either this or document, not both)
            image_resolution_dpi: Optional image resolution DPI.
            browser_canvas: Optional browser canvas size.
            temperature: Model temperature setting (0-1)
            modality: Modality of the document (e.g., native)
            reasoning_effort: The effort level for the model to reason about the input data.
            n_consensus: Number of consensus extractions to perform (default: 1 which computes a single extraction and the likelihoods comes from the model logprobs)
            idempotency_key: Idempotency key for request
            store: Whether to store the document in the Retab database

        Returns:
            Generator[RetabParsedChatCompletion]: Stream of parsed responses
        Raises:
            ValueError: If neither document nor documents is provided, or if both are provided
            HTTPException: If the request fails
        Usage:
        ```python
        # Single document
        with retab.documents.extractions.stream(json_schema, model, document=document) as stream:
            for response in stream:
                print(response)

        # Multiple documents
        with retab.documents.extractions.stream(json_schema, model, documents=[doc1, doc2]) as stream:
            for response in stream:
                print(response)
        ```
        """
        request = self.prepare_extraction(
            json_schema=json_schema,
            document=document,
            documents=documents,
            image_resolution_dpi=image_resolution_dpi,
            browser_canvas=browser_canvas,
            model=model,
            temperature=temperature,
            modality=modality,
            reasoning_effort=reasoning_effort,
            stream=True,
            n_consensus=n_consensus,
            store=store,
            idempotency_key=idempotency_key,
        )
        schema = Schema(json_schema=load_json_schema(json_schema))

        # Request the stream and return a context manager
        ui_parsed_chat_completion_cum_chunk: RetabParsedChatCompletionChunk | None = None
        # Initialize the RetabParsedChatCompletion object
        ui_parsed_completion: RetabParsedChatCompletion = RetabParsedChatCompletion(
            id="",
            created=0,
            model="",
            object="chat.completion",
            likelihoods={},
            choices=[
                RetabParsedChoice(
                    index=0,
                    message=ParsedChatCompletionMessage(content="", role="assistant"),
                    finish_reason=None,
                    logprobs=None,
                )
            ],
        )
        for chunk_json in self._client._prepared_request_stream(request):
            if not chunk_json:
                continue
            ui_parsed_chat_completion_cum_chunk = RetabParsedChatCompletionChunk.model_validate(chunk_json).chunk_accumulator(ui_parsed_chat_completion_cum_chunk)
            # Basic stuff
            ui_parsed_completion.id = ui_parsed_chat_completion_cum_chunk.id
            ui_parsed_completion.created = ui_parsed_chat_completion_cum_chunk.created
            ui_parsed_completion.model = ui_parsed_chat_completion_cum_chunk.model
            # Update the ui_parsed_completion object
            parsed = unflatten_dict(ui_parsed_chat_completion_cum_chunk.choices[0].delta.flat_parsed)
            likelihoods = unflatten_dict(ui_parsed_chat_completion_cum_chunk.choices[0].delta.flat_likelihoods)
            ui_parsed_completion.choices[0].message.content = json.dumps(parsed)
            ui_parsed_completion.choices[0].message.parsed = parsed
            ui_parsed_completion.likelihoods = likelihoods

            yield maybe_parse_to_pydantic(schema, ui_parsed_completion, allow_partial=True)

        # change the finish_reason to stop
        ui_parsed_completion.choices[0].finish_reason = "stop"
        yield maybe_parse_to_pydantic(schema, ui_parsed_completion)

    def log(
        self,
        document: Path | str | IOBase | HttpUrl | None,
        json_schema: dict[str, Any],
        model: str,
        temperature: float,
        completion: Any | None = None,
        # The messages can be provided in different formats, we will convert them to the Retab-compatible format
        messages: list[ChatCompletionRetabMessage] | None = None,
        openai_messages: list[ChatCompletionMessageParam] | None = None,
        anthropic_messages: list[MessageParam] | None = None,
        anthropic_system_prompt: str | None = None,
        # New fields for the Responses API
        openai_responses_input: list[ResponseInputItemParam] | None = None,
        openai_responses_output: Response | None = None,
    ) -> None:
        request = self.prepare_log_extraction(
            document=document,
            json_schema=json_schema,
            model=model,
            temperature=temperature,
            completion=completion,
            messages=messages,
            openai_messages=openai_messages,
            anthropic_messages=anthropic_messages,
            anthropic_system_prompt=anthropic_system_prompt,
            openai_responses_input=openai_responses_input,
            openai_responses_output=openai_responses_output,
        )
        return self._client._prepared_request(request)


class AsyncExtractions(AsyncAPIResource, BaseExtractionsMixin):
    """Extraction API wrapper for asynchronous usage."""

    async def parse(
        self,
        json_schema: dict[str, Any] | Path | str,
        model: str,
        document: Path | str | IOBase | HttpUrl | None = None,
        documents: list[Path | str | IOBase | HttpUrl] | None = None,
        image_resolution_dpi: int = FieldUnset,
        browser_canvas: BrowserCanvas = FieldUnset,
        temperature: float = FieldUnset,
        modality: Modality = FieldUnset,
        reasoning_effort: ChatCompletionReasoningEffort = FieldUnset,
        n_consensus: int = FieldUnset,
        idempotency_key: str | None = None,
        store: bool = False,
    ) -> RetabParsedChatCompletion:
        """
        Extract structured data from one or more documents asynchronously.

        Args:
            json_schema: JSON schema defining the expected data structure.
            model: The AI model to use.
            document: Single document to process (use either this or documents, not both)
            documents: List of documents to process (use either this or document, not both)
            image_resolution_dpi: Optional image resolution DPI.
            browser_canvas: Optional browser canvas size.
            temperature: Model temperature setting (0-1).
            modality: Modality of the document (e.g., native).
            reasoning_effort: The effort level for the model to reason about the input data.
            n_consensus: Number of consensus extractions to perform (default: 1 which computes a single extraction and the likelihoods comes from the model logprobs)
            idempotency_key: Idempotency key for request
            store: Whether to store the document in the Retab database
        Returns:
            RetabParsedChatCompletion: Parsed response from the API.
        Raises:
            ValueError: If neither document nor documents is provided, or if both are provided
        """
        request = self.prepare_extraction(
            json_schema=json_schema,
            document=document,
            documents=documents,
            image_resolution_dpi=image_resolution_dpi,
            browser_canvas=browser_canvas,
            model=model,
            temperature=temperature,
            modality=modality,
            reasoning_effort=reasoning_effort,
            stream=False,
            n_consensus=n_consensus,
            store=store,
            idempotency_key=idempotency_key,
        )
        response = await self._client._prepared_request(request)
        schema = Schema(json_schema=load_json_schema(json_schema))
        return maybe_parse_to_pydantic(schema, RetabParsedChatCompletion.model_validate(response))

    @as_async_context_manager
    async def stream(
        self,
        json_schema: dict[str, Any] | Path | str,
        model: str,
        document: Path | str | IOBase | HttpUrl | None = None,
        documents: list[Path | str | IOBase | HttpUrl] | None = None,
        image_resolution_dpi: int = FieldUnset,
        browser_canvas: BrowserCanvas = FieldUnset,
        temperature: float = FieldUnset,
        modality: Modality = FieldUnset,
        reasoning_effort: ChatCompletionReasoningEffort = FieldUnset,
        n_consensus: int = FieldUnset,
        idempotency_key: str | None = None,
        store: bool = False,
    ) -> AsyncGenerator[RetabParsedChatCompletion, None]:
        """
        Extract structured data from one or more documents asynchronously with streaming.

        Args:
            json_schema: JSON schema defining the expected data structure.
            model: The AI model to use.
            document: Single document to process (use either this or documents, not both)
            documents: List of documents to process (use either this or document, not both)
            image_resolution_dpi: Optional image resolution DPI.
            browser_canvas: Optional browser canvas size.
            temperature: Model temperature setting (0-1).
            modality: Modality of the document (e.g., native).
            reasoning_effort: The effort level for the model to reason about the input data.
            n_consensus: Number of consensus extractions to perform (default: 1 which computes a single extraction and the likelihoods comes from the model logprobs)
            idempotency_key: Idempotency key for request
            store: Whether to store the document in the Retab database
        Returns:
            AsyncGenerator[RetabParsedChatCompletion, None]: Stream of parsed responses.
        Raises:
            ValueError: If neither document nor documents is provided, or if both are provided

        Usage:
        ```python
        # Single document
        async with retab.documents.extractions.stream(json_schema, model, document=document) as stream:
            async for response in stream:
                print(response)

        # Multiple documents
        async with retab.documents.extractions.stream(json_schema, model, documents=[doc1, doc2]) as stream:
            async for response in stream:
                print(response)
        ```
        """
        request = self.prepare_extraction(
            json_schema=json_schema,
            document=document,
            documents=documents,
            image_resolution_dpi=image_resolution_dpi,
            browser_canvas=browser_canvas,
            model=model,
            temperature=temperature,
            modality=modality,
            reasoning_effort=reasoning_effort,
            stream=True,
            n_consensus=n_consensus,
            store=store,
            idempotency_key=idempotency_key,
        )
        schema = Schema(json_schema=load_json_schema(json_schema))
        ui_parsed_chat_completion_cum_chunk: RetabParsedChatCompletionChunk | None = None
        # Initialize the RetabParsedChatCompletion object
        ui_parsed_completion: RetabParsedChatCompletion = RetabParsedChatCompletion(
            id="",
            created=0,
            model="",
            object="chat.completion",
            likelihoods={},
            choices=[
                RetabParsedChoice(
                    index=0,
                    message=ParsedChatCompletionMessage(content="", role="assistant"),
                    finish_reason=None,
                    logprobs=None,
                )
            ],
        )

        async for chunk_json in self._client._prepared_request_stream(request):
            if not chunk_json:
                continue
            ui_parsed_chat_completion_cum_chunk = RetabParsedChatCompletionChunk.model_validate(chunk_json).chunk_accumulator(ui_parsed_chat_completion_cum_chunk)
            # Basic stuff
            ui_parsed_completion.id = ui_parsed_chat_completion_cum_chunk.id
            ui_parsed_completion.created = ui_parsed_chat_completion_cum_chunk.created
            ui_parsed_completion.model = ui_parsed_chat_completion_cum_chunk.model

            # Update the ui_parsed_completion object
            parsed = unflatten_dict(ui_parsed_chat_completion_cum_chunk.choices[0].delta.flat_parsed)
            likelihoods = unflatten_dict(ui_parsed_chat_completion_cum_chunk.choices[0].delta.flat_likelihoods)
            ui_parsed_completion.choices[0].message.content = json.dumps(parsed)
            ui_parsed_completion.choices[0].message.parsed = parsed
            ui_parsed_completion.likelihoods = likelihoods

            yield maybe_parse_to_pydantic(schema, ui_parsed_completion, allow_partial=True)

        # change the finish_reason to stop
        ui_parsed_completion.choices[0].finish_reason = "stop"
        yield maybe_parse_to_pydantic(schema, ui_parsed_completion)

    async def log(
        self,
        document: Path | str | IOBase | HttpUrl | None,
        json_schema: dict[str, Any],
        model: str,
        temperature: float,
        completion: Any | None = None,
        # The messages can be provided in different formats, we will convert them to the Retab-compatible format
        messages: list[ChatCompletionRetabMessage] | None = None,
        openai_messages: list[ChatCompletionMessageParam] | None = None,
        anthropic_messages: list[MessageParam] | None = None,
        anthropic_system_prompt: str | None = None,
        # New fields for the Responses API
        openai_responses_input: list[ResponseInputItemParam] | None = None,
        openai_responses_output: Response | None = None,
    ) -> None:
        request = self.prepare_log_extraction(
            document=document,
            json_schema=json_schema,
            model=model,
            temperature=temperature,
            completion=completion,
            messages=messages,
            openai_messages=openai_messages,
            anthropic_messages=anthropic_messages,
            anthropic_system_prompt=anthropic_system_prompt,
            openai_responses_input=openai_responses_input,
            openai_responses_output=openai_responses_output,
        )
        return await self._client._prepared_request(request)



from typing import Optional, TypeVar, Union

from openai.types.responses.response import Response
from openai.types.responses.response_input_param import ResponseInputParam
from openai.types.responses.response_text_config_param import ResponseTextConfigParam
from openai.types.shared_params.reasoning import Reasoning
from pydantic import BaseModel

from ..._resource import AsyncAPIResource, SyncAPIResource
from ...utils.ai_models import assert_valid_model_extraction
from ...types.completions import RetabChatResponseCreateRequest
from ...types.documents.extractions import UiResponse
from ...types.schemas.object import Schema
from ...types.standards import PreparedRequest

T = TypeVar("T", bound=BaseModel)


class BaseResponsesMixin:
    def prepare_create(
        self,
        model: str,
        input: Union[str, ResponseInputParam],
        text: ResponseTextConfigParam,
        temperature: float = 0,
        reasoning: Optional[Reasoning] = None,
        n_consensus: int = 1,
        instructions: Optional[str] = None,
        idempotency_key: Optional[str] = None,
    ) -> PreparedRequest:
        """
        Prepare a request for the Responses API create method.
        """
        assert_valid_model_extraction(model)

        text_format = text.get("format", None)
        assert text_format is not None, "text.format is required"
        json_schema = text_format.get("schema", None)
        assert json_schema is not None, "text.format.schema is required"

        schema_obj = Schema(json_schema=json_schema)

        if instructions is None:
            instructions = schema_obj.developer_system_prompt

        # Create the request object based on the RetabChatResponseCreateRequest model
        request = RetabChatResponseCreateRequest(
            model=model,
            input=input,
            temperature=temperature,
            stream=False,
            reasoning=reasoning,
            n_consensus=n_consensus,
            text={"format": {"type": "json_schema", "name": schema_obj.id, "schema": schema_obj.inference_json_schema, "strict": True}},
            instructions=instructions,
        )

        return PreparedRequest(method="POST", url="/v1/responses", data=request.model_dump(mode="json", exclude_unset=True), idempotency_key=idempotency_key)

    def prepare_parse(
        self,
        model: str,
        input: Union[str, ResponseInputParam],
        text_format: type[BaseModel],
        temperature: float = 0,
        reasoning: Optional[Reasoning] = None,
        n_consensus: int = 1,
        instructions: Optional[str] = None,
        idempotency_key: Optional[str] = None,
    ) -> PreparedRequest:
        """
        Prepare a request for the Responses API parse method.
        """

        assert_valid_model_extraction(model)

        schema_obj = Schema(pydantic_model=text_format)

        if instructions is None:
            instructions = schema_obj.developer_system_prompt

        # Create the request object based on the RetabChatResponseCreateRequest model
        request = RetabChatResponseCreateRequest(
            model=model,
            input=input,
            temperature=temperature,
            stream=False,
            reasoning=reasoning,
            n_consensus=n_consensus,
            text={"format": {"type": "json_schema", "name": schema_obj.id, "schema": schema_obj.inference_json_schema, "strict": True}},
            instructions=instructions,
        )

        return PreparedRequest(
            method="POST",
            url="/v1/responses",
            data=request.model_dump(),
            idempotency_key=idempotency_key,
        )


class Responses(SyncAPIResource, BaseResponsesMixin):
    """Retab Responses API compatible with OpenAI Responses API"""

    def create(
        self,
        model: str,
        input: Union[str, ResponseInputParam],
        text: ResponseTextConfigParam,
        temperature: float = 0,
        reasoning: Optional[Reasoning] = None,
        n_consensus: int = 1,
        instructions: Optional[str] = None,
        idempotency_key: Optional[str] = None,
    ) -> Response:
        """
        Create a completion using the Retab API with OpenAI Responses API compatible interface.

        Args:
            model: The model to use
            input: The input text or message array
            temperature: Model temperature setting (0-1)
            reasoning: The effort level for the model to reason about the input data
            n_consensus: Number of consensus models to use
            text: The response format configuration
            instructions: Optional system instructions
            idempotency_key: Idempotency key for request
        Returns:
            Response: OpenAI Responses API compatible response
        """
        request = self.prepare_create(
            model=model,
            input=input,
            temperature=temperature,
            reasoning=reasoning,
            text=text,
            instructions=instructions,
            n_consensus=n_consensus,
            idempotency_key=idempotency_key,
        )

        result = self._client._prepared_request(request)
        response = UiResponse.model_validate(result)

        return response

    def parse(
        self,
        model: str,
        input: Union[str, ResponseInputParam],
        text_format: type[T],
        temperature: float = 0,
        reasoning: Optional[Reasoning] = None,
        n_consensus: int = 1,
        instructions: Optional[str] = None,
        idempotency_key: Optional[str] = None,
    ) -> Response:
        """
        Parse content using the Retab API with OpenAI Responses API compatible interface.

        Args:
            model: The model to use
            input: The input text or message array
            text_format: The Pydantic model defining the expected output format
            temperature: Model temperature setting (0-1)
            reasoning_effort: The effort level for the model to reason about the input data
            n_consensus: Number of consensus models to use
            instructions: Optional system instructions
            idempotency_key: Idempotency key for request

        Returns:
            Response: OpenAI Responses API compatible response with parsed content
        """
        request = self.prepare_parse(
            model=model,
            input=input,
            temperature=temperature,
            reasoning=reasoning,
            text_format=text_format,
            instructions=instructions,
            n_consensus=n_consensus,
            idempotency_key=idempotency_key,
        )

        result = self._client._prepared_request(request)
        response = UiResponse.model_validate(result)

        return response


class AsyncResponses(AsyncAPIResource, BaseResponsesMixin):
    """Retab Responses API compatible with OpenAI Responses API for async usage"""

    async def create(
        self,
        model: str,
        input: Union[str, ResponseInputParam],
        text: ResponseTextConfigParam,
        temperature: float = 0,
        reasoning: Optional[Reasoning] = None,
        n_consensus: int = 1,
        instructions: Optional[str] = None,
        idempotency_key: Optional[str] = None,
    ) -> UiResponse:
        """
        Create a completion using the Retab API asynchronously with OpenAI Responses API compatible interface.

        Args:
            model: The model to use
            input: The input text or message array
            text: The response format configuration
            temperature: Model temperature setting (0-1)
            reasoning: The effort level for the model to reason about the input data
            n_consensus: Number of consensus models to use
            instructions: Optional system instructions
            idempotency_key: Idempotency key for request

        Returns:
            Response: OpenAI Responses API compatible response
        """
        request = self.prepare_create(
            model=model,
            input=input,
            temperature=temperature,
            reasoning=reasoning,
            text=text,
            instructions=instructions,
            n_consensus=n_consensus,
            idempotency_key=idempotency_key,
        )

        result = await self._client._prepared_request(request)
        response = UiResponse.model_validate(result)
        return response

    async def parse(
        self,
        model: str,
        input: Union[str, ResponseInputParam],
        text_format: type[BaseModel],
        temperature: float = 0,
        reasoning: Optional[Reasoning] = None,
        n_consensus: int = 1,
        instructions: Optional[str] = None,
        idempotency_key: Optional[str] = None,
    ) -> UiResponse:
        """
        Parse content using the Retab API asynchronously with OpenAI Responses API compatible interface.

        Args:
            model: The model to use
            input: The input text or message array
            text_format: The Pydantic model defining the expected output format
            temperature: Model temperature setting (0-1)
            reasoning: The effort level for the model to reason about the input data
            n_consensus: Number of consensus models to use
            instructions: Optional system instructions
            idempotency_key: Idempotency key for request

        Returns:
            Response: OpenAI Responses API compatible response with parsed content
        """
        request = self.prepare_parse(
            model=model,
            input=input,
            temperature=temperature,
            reasoning=reasoning,
            text_format=text_format,
            instructions=instructions,
            n_consensus=n_consensus,
            idempotency_key=idempotency_key,
        )

        result = await self._client._prepared_request(request)
        response = UiResponse.model_validate(result)
        return response
